from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from functools import wraps
from types import EllipsisType, FunctionType, MethodType, ModuleType, NoneType, NotImplementedType, UnionType
from typing import TYPE_CHECKING, Annotated, Final, Literal, TypeVar, Union, get_origin, overload

if TYPE_CHECKING:
    from sonolus.script.internal.value import Value


@overload
def meta_fn[T: Callable](fn: T) -> T: ...


@overload
def meta_fn[T: Callable](show_in_stack: bool) -> Callable[[T], T]: ...


def meta_fn(fn=None, *, show_in_stack: bool = True):
    """Marks a function as a meta function to be called directly without the AST visitor.

    This can also improve performance in some cases by avoiding the overhead of the AST visitor.
    """

    # noinspection PyShadowingNames
    def decorator(fn):
        from sonolus.backend.utils import get_function

        base_fn = fn
        while hasattr(base_fn, "__wrapped__"):
            base_fn = base_fn.__wrapped__

        function_name = getattr(base_fn, "__name__", "<unnamed>")
        module = getattr(base_fn, "__module__", None)
        if module is not None:
            qualified_name = f"{module}.{getattr(base_fn, '__qualname__', function_name)}"
        else:
            qualified_name = f"<unknown>.{getattr(base_fn, '__qualname__', function_name)}"
        source_file, node = get_function(base_fn)

        @wraps(fn)
        def wrapper(*args, **kwargs):
            from sonolus.backend.visitor import mark_start
            from sonolus.script.internal.context import ctx

            if ctx():
                completion_timer = mark_start(qualified_name)
                debug_stack = ctx().callback_state.debug_stack
                try:
                    if show_in_stack:
                        debug_stack.append(f'File "{source_file}", line {node.lineno}, in {function_name}')
                    return fn(*args, **kwargs)
                finally:
                    if show_in_stack:
                        debug_stack.pop()
                    completion_timer()
            return fn(*args, **kwargs)

        wrapper._meta_fn_ = True
        return wrapper

    if fn is None:
        return decorator
    return decorator(fn)


# To indicate this was used for performance reasons rather than functionality
perf_meta_fn = meta_fn


def validate_value[T](value: T) -> Value | T:
    from sonolus.script.internal.value import Value

    if isinstance(value, Value):
        return value

    from sonolus.script.globals import _GlobalPlaceholder
    from sonolus.script.internal.constant import BasicConstantValue, TypingSpecialFormConstant
    from sonolus.script.internal.dict_impl import DictImpl
    from sonolus.script.internal.generic import PartialGeneric
    from sonolus.script.internal.tuple_impl import TupleImpl
    from sonolus.script.num import Num

    try:
        # Unfortunately this is called during import, so this may fail
        from sonolus.script.internal.builtin_impls import BUILTIN_IMPLS

        if id(value) in BUILTIN_IMPLS:
            return validate_value(BUILTIN_IMPLS[id(value)])
    except ImportError:
        pass

    if hasattr(value, "_init_") and callable(value._init_):
        try:
            value._init_()
        except Exception as e:
            raise RuntimeError(f"Error initializing value {value}: {e}") from e

    match value:
        case int() | float() | bool():
            return Num._accept_(value)
        case Enum():
            return validate_value(value.value)
        case type():
            if value in {int, float, bool}:
                return BasicConstantValue.of(Num)
            return BasicConstantValue.of(value)
        case tuple():
            return TupleImpl._accept_(value)
        case dict():
            return DictImpl._accept_(value)
        case set() | frozenset():
            from sonolus.script.containers import FrozenNumSet

            return FrozenNumSet.of(*value)
        case (
            PartialGeneric()
            | TypeVar()
            | FunctionType()
            | MethodType()
            | str()
            | ModuleType()
            | NoneType()
            | NotImplementedType()
            | EllipsisType()
            | super()
        ):
            return BasicConstantValue.of(value)
        case special_form if value == Literal or value == Annotated or value == Union:  # noqa: PLR1714, SIM109
            return TypingSpecialFormConstant.of(special_form)
        case other_type if get_origin(value) in {Literal, Annotated, UnionType, Final, tuple, type}:
            return BasicConstantValue.of(other_type)
        case _GlobalPlaceholder():
            return value.get()
        case comptime_value if getattr(comptime_value, "_is_comptime_value_", False):
            return BasicConstantValue.of(comptime_value)
        case _:
            raise TypeError(f"Unsupported value: {value!r}")



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from tests.script.conftest import run_and_validate
from tests.script.test_record import Pair


@given(n=st.integers(min_value=0, max_value=100))
def test_basic_range_iteration(n):
    def fn():
        total = 0
        for i in range(n):
            total += i
        return total

    expected = sum(range(n))
    result = run_and_validate(fn)
    assert result == expected


@given(start=st.integers(-100, 100), stop=st.integers(-100, 100))
def test_range_iteration_with_start(start, stop):
    def fn():
        total = 0
        for i in range(start, stop):
            total += i
        return total

    expected = sum(range(start, stop))
    result = run_and_validate(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_range_iteration_with_step(start, stop, step):
    def fn():
        total = 0
        for i in range(start, stop, step):
            total += i
        return total

    expected = sum(range(start, stop, step))
    result = run_and_validate(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, -1),
)
def test_range_iteration_with_negative_step(start, stop, step):
    def fn():
        total = 0
        for i in range(start, stop, step):
            total += i
        return total

    expected = sum(range(start, stop, step))
    result = run_and_validate(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
    value=st.integers(-200, 200),
)
def test_range_contains(start, stop, step, value):
    def fn():
        return value in range(start, stop, step)

    expected = value in range(start, stop, step)
    result = run_and_validate(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_range_size(start, stop, step):
    def fn():
        return len(range(start, stop, step))

    expected = len(range(start, stop, step))
    result = run_and_validate(fn)
    assert result == expected


@given(
    start1=st.integers(-100, 100),
    stop1=st.integers(-100, 100),
    step1=st.integers(-10, 10).filter(lambda x: x != 0),
    start2=st.integers(-100, 100),
    stop2=st.integers(-100, 100),
    step2=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_range_equality(start1, stop1, step1, start2, stop2, step2):
    def fn():
        a = range(start1, stop1, step1)
        b = range(start2, stop2, step2)
        return Pair(a == b, b == a)

    range_a = range(start1, stop1, step1)
    range_b = range(start2, stop2, step2)
    expected = Pair(range_a == range_b, range_b == range_a)
    result = run_and_validate(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_identical_range_equality(start, stop, step):
    def fn():
        a = range(start, stop, step)
        b = range(start, stop, step)
        return Pair(a == b, b == a)

    expected = Pair(True, True)
    result = run_and_validate(fn)
    assert result == expected


def test_range_negative_indexing():
    def fn():
        r = range(10, 50, 5)

        return Array(
            r[-1],
            r[-2],
            r[-3],
            r[-4],
            r[-5],
            r[-6],
            r[-7],
            r[-8],
        )

    assert list(run_and_validate(fn)) == [45, 40, 35, 30, 25, 20, 15, 10]


@given(
    start=st.integers(-50, 50),
    stop=st.integers(-50, 50),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_range_negative_positive_indexing_equivalence(start, stop, step):
    length = len(range(start, stop, step))

    def fn():
        r = range(start, stop, step)

        if length == 0:
            return Array(True)

        results = VarArray[bool, length].new()
        for i in range(1, length + 1):
            if i <= length:
                results.append(r[-i] == r[length - i])

        return results

    assert all(run_and_validate(fn))


def test_range_indexing_with_arrays():
    def fn():
        r = range(5, 25, 3)
        results = VarArray[int, 7].new()

        for i in range(len(r)):
            results.append(r[i])

        return results

    assert list(run_and_validate(fn)) == [5, 8, 11, 14, 17, 20, 23]


def test_range_truthiness_empty():
    def fn():
        x = range(0)
        return 1 if x else 0

    assert run_and_validate(fn) == 0


def test_range_truthiness_non_empty():
    def fn():
        x = range(5)
        return 1 if x else 0

    assert run_and_validate(fn) == 1



from __future__ import annotations

from typing import cast

from pydori.lib.layout import get_note_y, preempt_time
from pydori.lib.note import (
    NoteKind,
    destroy_particle,
    draw_note,
    draw_note_head,
    get_note_bucket,
    init_note_life,
    play_note_particle,
    schedule_hold_sfx,
    schedule_note_sfx,
    update_hold_particle,
)
from pydori.lib.options import Options
from pydori.lib.streams import Streams
from sonolus.script.archetype import (
    EntityRef,
    StandardImport,
    WatchArchetype,
    entity_data,
    entity_memory,
    imported,
    shared_memory,
)
from sonolus.script.bucket import Judgment
from sonolus.script.particle import ParticleHandle
from sonolus.script.runtime import is_replay, is_skip, scaled_time, time
from sonolus.script.timing import beat_to_time, time_to_scaled_time


class WatchNote(WatchArchetype):
    """Common archetype for notes."""

    lane: float = imported()
    beat: StandardImport.BEAT = imported()
    direction: int = imported()
    prev_ref: EntityRef[WatchNote] = imported()
    next_ref: EntityRef[WatchNote] = imported()

    target_time: float = entity_data()
    target_scaled_time: float = entity_data()
    start_scaled_time: float = entity_data()
    end_scaled_time: float = entity_data()
    head_ref: EntityRef[WatchNote] = entity_data()
    end_ref: EntityRef[WatchNote] = entity_data()

    _hold_lane: float = shared_memory()

    end_time: float = imported()
    judgment: StandardImport.JUDGMENT = imported()
    accuracy: StandardImport.ACCURACY = imported()

    def preprocess(self):
        if Options.mirror:
            self.lane = -self.lane
            self.direction = -self.direction

        self.target_time = beat_to_time(self.beat)
        self.target_scaled_time = time_to_scaled_time(self.target_time)
        self.start_scaled_time = self.target_scaled_time - preempt_time()
        self.end_scaled_time = time_to_scaled_time(self.end_time)
        self.result.bucket = get_note_bucket(self.kind)

        self.result.target_time = self.target_time

        self.head_ref = self.ref()
        while self.head.has_prev:
            self.head_ref = self.head.prev_ref

        self.end_ref = self.ref()
        while self.end.has_next:
            self.end_ref = self.end.next_ref

        if is_replay():
            if self.judgment != Judgment.MISS:
                schedule_note_sfx(self.kind, self.judgment, self.end_time)
            self.result.bucket_value = self.accuracy * 1000
        else:
            self.judgment = Judgment.PERFECT
            schedule_note_sfx(self.kind, Judgment.PERFECT, self.target_time)

        if self.has_prev and not self.has_next:
            WatchHoldManager.spawn(head_ref=self.head_ref, end_ref=self.end_ref)
            self.schedule_hold_sfx()

    def spawn_time(self) -> float:
        return self.start_scaled_time

    def despawn_time(self) -> float:
        if is_replay():
            return self.end_scaled_time
        else:
            return self.target_scaled_time

    def update_parallel(self):
        draw_note(
            self.kind,
            self.lane,
            self.y,
            self.direction,
        )

    def terminate(self):
        if is_skip():
            return
        if self.judgment != Judgment.MISS:
            play_note_particle(self.kind, self.lane, self.direction)

    def schedule_hold_sfx(self):
        if is_replay():
            was_active = False
            start_time = 0
            for input_time, is_active in Streams.hold_activity[self.head.index].iter_items_from(-10):
                if input_time < self.head.target_time:
                    # This is before the head meets the judge line, so it's irrelevant.
                    continue
                if input_time > self.end.target_time:
                    # This is after the end, so we can stop checking.
                    break
                if is_active and not was_active:
                    # Record the start time of a timespan where the hold is active.
                    start_time = input_time
                    was_active = True
                if not is_active and was_active:
                    # An active timespan has ended, so we can schedule the hold sfx over it.
                    schedule_hold_sfx(start_time, input_time)
                    was_active = False
            if was_active:
                # If the last recorded value was that the hold was active, we can assume it was active until the end,
                # and schedule the sfx from the current start time to the end of the hold.
                schedule_hold_sfx(start_time, self.end.target_time)
        else:
            schedule_hold_sfx(self.head.target_time, self.end.target_time)

    @classmethod
    def global_preprocess(cls):
        init_note_life(cls)

    @property
    def kind(self) -> NoteKind:
        return cast(NoteKind, self.key)

    @property
    def y(self) -> float:
        return get_note_y(scaled_time(), self.target_scaled_time)

    @property
    def has_prev(self) -> bool:
        return self.prev_ref.index > 0

    @property
    def prev(self) -> WatchNote:
        return self.prev_ref.get()

    @property
    def has_next(self) -> bool:
        return self.next_ref.index > 0

    @property
    def next(self) -> WatchNote:
        return self.next_ref.get()

    @property
    def head(self) -> WatchNote:
        return self.head_ref.get()

    @property
    def end(self) -> WatchNote:
        return self.end_ref.get()

    @property
    def has_active_touch(self) -> bool:
        if is_replay():
            return Streams.hold_activity[self.head.index].get_previous_inclusive(time())
        else:
            return True

    @property
    def hold_lane(self) -> float:
        return self.head._hold_lane

    @hold_lane.setter
    def hold_lane(self, value: float):
        self.head._hold_lane = value


WatchTapNote = WatchNote.derive("Tap", is_scored=True, key=NoteKind.TAP)
WatchFlickNote = WatchNote.derive("Flick", is_scored=True, key=NoteKind.FLICK)
WatchDirectionalFlickNote = WatchNote.derive("DirectionalFlick", is_scored=True, key=NoteKind.DIRECTIONAL_FLICK)
WatchHoldHeadNote = WatchNote.derive("HoldHead", is_scored=True, key=NoteKind.HOLD_HEAD)
WatchHoldTickNote = WatchNote.derive("HoldTick", is_scored=True, key=NoteKind.HOLD_TICK)
WatchHoldAnchorNote = WatchNote.derive("HoldAnchor", is_scored=False, key=NoteKind.HOLD_ANCHOR)
WatchHoldEndNote = WatchNote.derive("HoldEnd", is_scored=True, key=NoteKind.HOLD_END)

ALL_WATCH_NOTE_TYPES = (
    WatchTapNote,
    WatchFlickNote,
    WatchDirectionalFlickNote,
    WatchHoldHeadNote,
    WatchHoldTickNote,
    WatchHoldAnchorNote,
    WatchHoldEndNote,
)


class WatchHoldManager(WatchArchetype):
    """Manages the particle of a hold note.

    Unlike in play mode, the sfx is scheduled ahead of time so it's not handled here.
    """

    name = "HoldManager"

    head_ref: EntityRef[WatchNote] = entity_memory()
    end_ref: EntityRef[WatchNote] = entity_memory()

    particle: ParticleHandle = entity_memory()

    def spawn_time(self) -> float:
        return self.head.target_scaled_time

    def despawn_time(self) -> float:
        return self.end.target_scaled_time

    def update_parallel(self):
        if is_skip():
            destroy_particle(self.particle)
        if self.head.has_active_touch:
            draw_note_head(self.head.hold_lane)
            update_hold_particle(self.particle, self.head.hold_lane)
        else:
            destroy_particle(self.particle)

    def terminate(self):
        destroy_particle(self.particle)

    @property
    def head(self) -> WatchNote:
        return self.head_ref.get()

    @property
    def end(self) -> WatchNote:
        return self.end_ref.get()



import ast
import inspect
from collections.abc import Callable
from functools import cache
from pathlib import Path


@cache
def get_function(fn: Callable) -> tuple[str, ast.FunctionDef]:
    # This preserves both line number and column number in the returned node
    source_file = inspect.getsourcefile(fn)
    _, start_line = inspect.getsourcelines(fn)
    base_tree = get_tree_from_file(source_file)
    try:
        return source_file, find_function(base_tree, start_line)
    except ValueError:
        raise ValueError(f"Function {fn} not found in source file {source_file}") from None


@cache
def get_tree_from_file(file: str | Path) -> ast.Module:
    return ast.parse(Path(file).read_text(encoding="utf-8"))


class FindFunction(ast.NodeVisitor):
    def __init__(self, line):
        self.line = line
        self.results: list[ast.FunctionDef | ast.Lambda] = []
        self.current_fn = None

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.results.append(node)
        outer_fn = self.current_fn
        self.current_fn = node
        self.generic_visit(node)
        self.current_fn = outer_fn

    def visit_Lambda(self, node: ast.Lambda):
        self.results.append(node)
        outer_fn = self.current_fn
        self.current_fn = node
        self.generic_visit(node)
        self.current_fn = outer_fn

    # Visitors have high overhead, so we detect generators here rather than in a separate pass.

    def visit_Yield(self, node):
        self.current_fn.has_yield = True

    def visit_YieldFrom(self, node):
        self.current_fn.has_yield = True


@cache
def get_functions(tree: ast.Module) -> list[ast.FunctionDef | ast.Lambda]:
    visitor = FindFunction(0)
    visitor.visit(tree)
    return visitor.results


def find_function(tree: ast.Module, line: int):
    for node in get_functions(tree):
        if node.lineno == line or (
            isinstance(node, ast.FunctionDef)
            and node.decorator_list
            and (node.decorator_list[0].end_lineno <= line <= node.lineno)
        ):
            return node
    raise ValueError("Function not found")


class ScanWrites(ast.NodeVisitor):
    def __init__(self):
        self.writes = []

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store | ast.Delete):
            self.writes.append(node.id)


def scan_writes(node: ast.AST) -> set[str]:
    visitor = ScanWrites()
    visitor.visit(node)
    return set(visitor.writes)



from pydori.lib.effect import Effects
from pydori.lib.particle import Particles
from pydori.lib.skin import Skin
from pydori.tutorial.instructions import InstructionIcons, Instructions
from pydori.tutorial.navigate import navigate
from pydori.tutorial.preprocess import preprocess
from pydori.tutorial.update import update
from sonolus.script.engine import TutorialMode

tutorial_mode = TutorialMode(
    skin=Skin,
    effects=Effects,
    particles=Particles,
    instructions=Instructions,
    instruction_icons=InstructionIcons,
    preprocess=preprocess,
    navigate=navigate,
    update=update,
)



from __future__ import annotations

from pydori.lib.layout import Layout
from pydori.lib.note import destroy_particle, stop_looped_sfx
from sonolus.script.effect import LoopedEffectHandle
from sonolus.script.globals import level_memory
from sonolus.script.instruction import clear_instruction
from sonolus.script.interval import lerp, remap, remap_clamped
from sonolus.script.particle import ParticleHandle
from sonolus.script.record import Record
from sonolus.script.runtime import time


@level_memory
class PhaseState:
    start_time: float
    hold_particle: ParticleHandle
    hold_sfx: LoopedEffectHandle
    hold_was_accessed: bool
    prev_time: float


def update_start():
    """Update tutorial state at the beginning of each frame."""
    PhaseState.hold_was_accessed = False
    clear_instruction()


def update_end():
    """Update tutorial state at the end of each frame."""
    if not PhaseState.hold_was_accessed:
        destroy_particle(PhaseState.hold_particle)
        stop_looped_sfx(PhaseState.hold_sfx)
    PhaseState.prev_time = time()


def reset_phase():
    """Reset the tutorial phase time to zero."""
    PhaseState.start_time = time()


def get_hold_particle() -> ParticleHandle:
    """Mark that the hold particle was accessed and return it."""
    PhaseState.hold_was_accessed = True
    return PhaseState.hold_particle


def get_hold_sfx() -> LoopedEffectHandle:
    """Mark that the hold sound effect was accessed and return it."""
    PhaseState.hold_was_accessed = True
    return PhaseState.hold_sfx


def current_phase_time() -> PhaseTime:
    """Return the current phase timing information."""
    return PhaseTime(time() - PhaseState.start_time, PhaseState.prev_time - PhaseState.start_time)


class PhaseTime(Record):
    time: float
    prev_time: float

    def range(self, start: float, end: float, segments: int = 1) -> PhaseRange:
        return PhaseRange(self, start, end, segments)

    def first(self, end: float, repeats: int = 1) -> PhaseRange:
        return self.range(0, end * repeats, segments=repeats)

    def instant(self, timing: float) -> PhaseInstant:
        return PhaseInstant(self, timing)


class PhaseRange(Record):
    phase: PhaseTime
    start: float
    end: float
    segments: int

    @property
    def is_active(self):
        return self.start <= self.phase.time < self.end

    @property
    def progress(self) -> float:
        if self.segments == 1:
            return remap(self.start, self.end, 0, 1, self.phase.time)
        else:
            return remap_clamped(self.start, self.end, 0, 1, self.phase.time) * self.segments % 1

    @property
    def is_done(self):
        return self.phase.time >= self.end

    def next(self, duration: float, repeats: int = 1):
        return self.phase.range(self.end, self.end + duration * repeats, segments=repeats)

    def start_instant(self) -> PhaseInstant:
        return self.phase.instant(self.start)

    def end_instant(self) -> PhaseInstant:
        return self.phase.instant(self.end)

    def __bool__(self):
        return self.is_active


class PhaseInstant(Record):
    phase: PhaseTime
    timing: float

    @property
    def is_active(self):
        return self.phase.prev_time < self.timing <= self.phase.time

    @property
    def is_done(self):
        return self.phase.time >= self.timing and not self.is_active

    @property
    def is_upcoming(self):
        return self.phase.time < self.timing

    def __bool__(self):
        return self.is_active


def progress_to_y(p: float) -> float:
    """Interpolate a 0-1 progress value to a y-max to 0 y-coordinate.

    Progresses less than 0 and greater than 1 will result in y-coordinates
    above y-max or below 0, respectively.

    Args:
        p: A progress value where 1 is when the note meets the judgment line.

    Returns:
        The y-coordinate value that corresponds to the progress.
    """
    return lerp(Layout.note_y_max, 0, p)



import pytest
from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.debug import assert_true, debug_log
from sonolus.script.num import Num
from sonolus.script.record import Record
from tests.script.conftest import run_and_validate


class Simple(Record):
    value: float


class Generic[T](Record):
    value: T

    def __add__(self, other):
        return Generic(self.value + other.value)

    def __isub__(self, other):
        self.value = 123
        return self

    @property
    def prop(self):
        debug_log(1)
        return self.value

    @prop.setter
    def prop(self, value):
        debug_log(2)
        self.value = value

    def __iadd__(self, other):
        self.value += other.value
        debug_log(3)
        return self


class Pair[T, U](Record):
    first: T
    second: U


class ConcreteCompound(Record):
    a: Pair[Num, Num]
    b: Pair[Num, Num]


@given(a=st.floats(allow_nan=False, allow_infinity=False))
def test_simple_record(a):
    def fn():
        r = Simple(a)
        assert_true(r.value == a)
        return 1

    assert run_and_validate(fn) == 1


@given(a=st.floats(allow_nan=False, allow_infinity=False))
def test_generic_record_inference(a):
    def fn():
        r = Generic(a)
        assert_true(r.value == a)
        return 1

    assert run_and_validate(fn) == 1


@given(a=st.floats(allow_nan=False, allow_infinity=False))
def test_generic_record_explicit(a):
    def fn():
        r = Generic[Num](a)
        assert_true(r.value == a)
        return 1

    assert run_and_validate(fn) == 1


def test_concrete_compound_record():
    def fn():
        r = ConcreteCompound(Pair(1, 2), Pair(3, 4))
        r @= ConcreteCompound(Pair(5, 6), Pair(7, 8))
        return r

    assert run_and_validate(fn) == ConcreteCompound(Pair(5, 6), Pair(7, 8))


def test_record_rejects_wrong_type():
    with pytest.raises(TypeError):
        Simple(Simple(1))


def test_concrete_generic_record_rejects_wrong_type():
    with pytest.raises(TypeError):
        Generic[Simple](1)


def test_nested_generic():
    def fn():
        r = Generic[Generic[Num]](Generic[Num](1))
        assert_true(r.value.value == 1)
        r2 = Generic[Generic[Num]](Generic(2))
        assert_true(r2.value.value == 2)
        return 1

    assert run_and_validate(fn) == 1


def test_value_record_members_are_independent():
    def fn():
        v = 1
        r = Generic[Num](v)
        assert_true(r.value == v)
        r.value = 2
        assert_true(r.value == 2)
        assert_true(v == 1)
        return 1

    assert run_and_validate(fn) == 1


def test_reference_record_members_are_shared():
    def fn():
        inner = Generic[Num](1)
        outer = Generic(inner)
        assert_true(outer.value.value == 1)
        inner.value = 2
        assert_true(outer.value.value == 2)
        outer.value.value = 3
        assert_true(inner.value == 3)
        return 1

    assert run_and_validate(fn) == 1


def test_automatic_record_copy_assign():
    def fn():
        inner = Generic(1)
        other = Generic(2)
        outer = Generic(inner)
        outer.value = other
        assert_true(outer.value.value == 2)
        assert_true(inner.value == 2)
        return 1

    assert run_and_validate(fn) == 1


def test_record_operator_overloading():
    def fn():
        r1 = Generic(1)
        r2 = Generic(2)
        r = r1 + r2
        assert_true(r == Generic(3))
        return 1

    assert run_and_validate(fn) == 1


def test_record_inplace_operator_generation():
    def fn():
        r = Generic(1)
        r_ref = r
        r += Generic(2)
        assert_true(r == Generic(3))
        assert_true(r_ref == Generic(3))
        return 1

    assert run_and_validate(fn) == 1


def test_record_explicit_inplace_operator():
    def fn():
        r = Generic(1)
        r_ref = r
        r -= Generic(2)
        assert_true(r == Generic(123))
        assert_true(r_ref == Generic(123))
        return 1

    assert run_and_validate(fn) == 1


def test_record_equality():
    def fn():
        r1 = Pair(1, 2)
        r2 = Pair(1, 2)
        r3 = Pair(3, 4)
        r4 = Pair(Simple(1), Simple(2))
        r5 = Pair(Simple(1), Simple(2))
        r6 = Pair(Simple(3), Simple(4))

        assert_true(r1 == r2)
        assert_true(r1 != r3)
        assert_true(r4 == r5)
        assert_true(r4 != r6)

        return 1

    assert run_and_validate(fn) == 1


@given(
    a=st.integers(min_value=0, max_value=20),
    b=st.integers(min_value=0, max_value=20),
    c=st.integers(min_value=0, max_value=20),
    d=st.integers(min_value=0, max_value=20),
    e=st.integers(min_value=0, max_value=20),
    f=st.integers(min_value=0, max_value=20),
)
def test_record_modification_in_loop(a, b, c, d, e, f):
    # This serves as a test that optimizations don't break even for complex control flow
    def fn():
        r = Pair(a, b)
        while r.first < c:
            if r.first < d:
                for i in range(r.first, r.second):
                    if i == f:
                        break
                    r.second += i
                r @= Pair(a + 1, b - 1)
                break
            if r.first == e:
                r.first += 3
                continue
            r.first += 1
            r.second += 5
        return r

    run_and_validate(fn)


def test_record_property_getter():
    def fn():
        r = Generic(1)
        return r.prop

    assert run_and_validate(fn) == 1


def test_record_property_setter():
    def fn():
        r = Generic(1)
        r.prop = 2
        return r.value

    assert run_and_validate(fn) == 2


def test_record_augmented_property_setter():
    def fn():
        r = Generic(1)
        r.prop += 2
        return r.value

    assert run_and_validate(fn) == 3


def test_record_augmented_property_setter_with_explicit_implementation():
    def fn():
        r = Generic(Generic(1))
        r.prop += Generic(2)
        return r.value.value

    assert run_and_validate(fn) == 3


def test_type_of_generic_record():
    def fn():
        r = Generic(Pair(1, 2))
        t = type(r)
        return t(Pair(3, 4))

    assert run_and_validate(fn) == Generic(Pair(3, 4))



import pytest

from sonolus.script.array import Array
from sonolus.script.debug import debug_log
from sonolus.script.internal.error import CompilationError
from tests.script.conftest import run_and_validate, run_compiled


def test_simple_genexpr():
    def fn():
        gen = (i for i in range(3))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_with_expression():
    def fn():
        gen = (i * 2 for i in range(3))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_with_filter():
    def fn():
        gen = (i for i in range(10) if i % 2 == 0)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_with_array():
    def fn():
        arr = Array(1, 2, 3, 4, 5)
        gen = (x * 2 for x in arr)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_with_tuple():
    def fn():
        tup = (1, 2, 3, 4, 5)
        gen = (x * 2 for x in tup)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_nested_loops():
    def fn():
        gen = (i * j for i in range(3) for j in range(2))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_nested_loops_with_filter():
    def fn():
        gen = (i * j for i in range(3) for j in range(2) if i + j > 1)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_multiple_filters():
    def fn():
        gen = (i for i in range(20) if i % 2 == 0 if i % 3 == 0)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_with_builtin_sum():
    def fn():
        gen = (i for i in range(5))
        return sum(gen)

    assert run_and_validate(fn) == 10


def test_genexpr_with_builtin_max():
    def fn():
        gen = (i * 2 for i in range(5))
        return max(gen)

    assert run_and_validate(fn) == 8


def test_genexpr_with_builtin_min():
    def fn():
        gen = (i * 2 + 1 for i in range(5))
        return min(gen)

    assert run_and_validate(fn) == 1


def test_genexpr_with_any():
    def fn():
        gen = (i > 3 for i in range(10))
        return any(gen)

    assert run_and_validate(fn)


def test_genexpr_with_all():
    def fn():
        gen = (i >= 0 for i in range(5))
        return all(gen)

    assert run_and_validate(fn)


def test_genexpr_closure_variable():
    def fn():
        factor = 3
        gen = (i * factor for i in range(5))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_multiple_closure_variables():
    def fn():
        base = 2
        multiplier = 3
        gen = (i * multiplier + base for i in range(5))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_nested_arrays():
    def fn():
        arr = Array(Array(1, 2), Array(3, 4))
        gen = (x * 2 for sub_arr in arr for x in sub_arr)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_nested_tuples():
    def fn():
        tup = ((1, 2), (3, 4))
        gen = (x * 2 for sub_tup in tup for x in sub_tup)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_complex_expression():
    def fn():
        gen = (i**2 + 2 * i + 1 for i in range(5))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_early_break():
    def fn():
        gen = (i for i in range(100))
        for i, x in enumerate(gen):
            debug_log(x)
            if i >= 2:
                break

    run_and_validate(fn)


def test_genexpr_iterator_resume_after_break():
    def fn():
        gen = (i for i in range(10))

        for i, x in enumerate(gen):
            debug_log(x)
            if i >= 2:
                break

        for i, x in enumerate(gen):
            debug_log(x + 100)
            if i >= 1:
                break

    run_and_validate(fn)


def test_genexpr_empty():
    def fn():
        gen = (i for i in range(0))
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_filter_excludes_all():
    def fn():
        gen = (i for i in range(5) if i > 10)
        for x in gen:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_changing_closure_sequential():
    def fn():
        x = 0

        gen = (i + x for i in range(3))
        iterator = gen
        for val in iterator:
            debug_log(val)
            break

        x = 10
        for val in iterator:
            debug_log(val)

        return 0

    with pytest.raises(CompilationError, match=r"Binding 'x' has been modified.*"):
        run_compiled(fn)


def test_genexpr_changing_closure_loop():
    def fn():
        x = 0

        gen = (i + x for i in range(3))
        for val in gen:
            debug_log(val)
            x += 1

        return 0

    with pytest.raises(CompilationError, match=r"Binding 'x' has been modified.*"):
        run_compiled(fn)


def test_genexpr_of_gexpr():
    def fn():
        gen = ((i + j for j in range(4)) for i in range(5))
        for sub_gen in gen:
            for x in sub_gen:
                debug_log(x)
                break
            for x in sub_gen:
                debug_log(x + 100)

    run_and_validate(fn)


def test_genexpr_with_next():
    def fn():
        gen = (i * 2 for i in range(5))
        debug_log(next(gen))
        debug_log(next(gen))
        debug_log(next(gen))

    run_and_validate(fn)


def test_genexpr_with_iter():
    def fn():
        gen = (i * 2 for i in range(5))
        iterator = iter(gen)
        for x in iterator:
            debug_log(x)

    run_and_validate(fn)


def test_genexpr_eagerly_evaluates_first_item():
    def fn():
        def first():
            debug_log(1)
            return Array(1, 2, 3)

        def second():
            debug_log(2)
            return Array(4, 5, 6)

        debug_log(3)
        gen = (a + b for a in first() for b in second())
        debug_log(4)
        for x in gen:
            debug_log(x)
        debug_log(5)

    run_and_validate(fn)


def test_genexpr_eagerly_evaluates_first_item_tuples():
    def fn():
        def first():
            debug_log(1)
            return 1, 2, 3

        def second():
            debug_log(2)
            return 4, 5, 6

        debug_log(3)
        gen = (a + b for a in first() for b in second())
        debug_log(4)
        for x in gen:
            debug_log(x)
        debug_log(5)

    run_and_validate(fn)



from typing import Any, Self

from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.transient import TransientValue


class DictImpl(TransientValue):
    def __init__(self, value: dict):
        self.value = value

    @meta_fn
    def __getitem__(self, item):
        item = validate_value(item)
        if not item._is_py_():
            raise TypeError("Key must be a compile-time constant")
        item = item._as_py_()
        if item not in self.value:
            raise KeyError(item)
        return self.value[item]

    @meta_fn
    def __contains__(self, item):
        item = validate_value(item)
        if not item._is_py_():
            raise TypeError("Key must be a compile-time constant")
        item = item._as_py_()
        return item in self.value

    @meta_fn
    def __len__(self):
        return len(self.value)

    def __eq__(self, other):
        raise TypeError("Comparing dicts is not supported")

    __hash__ = None

    @meta_fn
    def __or__(self, other):
        if not isinstance(other, DictImpl):
            raise TypeError("Only dicts can be merged")
        return DictImpl({**self.value, **other.value})

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return isinstance(value, cls | dict)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept {value} as {cls.__name__}")
        if isinstance(value, cls):
            return value
        else:
            return cls({validate_value(k)._as_py_(): validate_value(v) for k, v in value.items()})

    def _is_py_(self) -> bool:
        return all(v._is_py_() for v in self.value.values())

    def _as_py_(self) -> Any:
        return {k: v._as_py_() for k, v in self.value.items()}


DictImpl.__name__ = "dict"
DictImpl.__qualname__ = "dict"



from __future__ import annotations

from math import cos, sin

from sonolus.script.internal.impl import perf_meta_fn
from sonolus.script.interval import lerp, remap
from sonolus.script.quad import Quad, QuadLike
from sonolus.script.record import Record
from sonolus.script.vec import Vec2


class Transform2d(Record):
    """A transformation matrix for 2D points.

    Usage:
        ```python
        Transform2d.new()
        ```
    """

    a00: float
    a01: float
    a02: float
    a10: float
    a11: float
    a12: float
    a20: float
    a21: float
    a22: float

    @classmethod
    def new(cls) -> Transform2d:
        """Create a new identity transform.

        Returns:
            A new identity transform.
        """
        return cls(
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
        )

    @perf_meta_fn
    def _compose(
        self,
        b00: float,
        b01: float,
        b02: float,
        b10: float,
        b11: float,
        b12: float,
        b20: float,
        b21: float,
        b22: float,
    ) -> Transform2d:
        # Multiply by b on the left (b @ a)
        a00 = self.a00 * b00 + self.a10 * b01 + self.a20 * b02
        a01 = self.a01 * b00 + self.a11 * b01 + self.a21 * b02
        a02 = self.a02 * b00 + self.a12 * b01 + self.a22 * b02
        a10 = self.a00 * b10 + self.a10 * b11 + self.a20 * b12
        a11 = self.a01 * b10 + self.a11 * b11 + self.a21 * b12
        a12 = self.a02 * b10 + self.a12 * b11 + self.a22 * b12
        a20 = self.a00 * b20 + self.a10 * b21 + self.a20 * b22
        a21 = self.a01 * b20 + self.a11 * b21 + self.a21 * b22
        a22 = self.a02 * b20 + self.a12 * b21 + self.a22 * b22
        return Transform2d(
            a00,
            a01,
            a02,
            a10,
            a11,
            a12,
            a20,
            a21,
            a22,
        )

    def translate(self, translation: Vec2, /) -> Transform2d:
        """Translate along the x and y axes and return a new transform.

        Args:
            translation: The translation vector.

        Returns:
            A new transform after translation.
        """
        return self._compose(
            1,
            0,
            translation.x,
            0,
            1,
            translation.y,
            0,
            0,
            1,
        )

    def scale(self, factor: Vec2, /) -> Transform2d:
        """Scale about the origin and return a new transform.

        Args:
            factor: The scale factor vector.

        Returns:
            A new transform after scaling.
        """
        return self._compose(
            factor.x,
            0,
            0,
            0,
            factor.y,
            0,
            0,
            0,
            1,
        )

    def scale_about(self, factor: Vec2, /, pivot: Vec2) -> Transform2d:
        """Scale about the pivot and return a new transform.

        Args:
            factor: The scale factor vector.
            pivot: The pivot point for scaling.

        Returns:
            A new transform after scaling.
        """
        return self.translate(-pivot).scale(factor).translate(pivot)

    def rotate(self, angle: float, /) -> Transform2d:
        """Rotate about the origin and return a new transform.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.

        Returns:
            A new transform after rotation.
        """
        c = cos(angle)
        s = sin(angle)
        return self._compose(
            c,
            -s,
            0,
            s,
            c,
            0,
            0,
            0,
            1,
        )

    def rotate_about(self, angle: float, /, pivot: Vec2) -> Transform2d:
        """Rotate about the pivot and return a new transform.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.
            pivot: The pivot point for rotation.

        Returns:
            A new transform after rotation.
        """
        return self.translate(-pivot).rotate(angle).translate(pivot)

    def shear_x(self, m: float, /) -> Transform2d:
        """Shear along the x-axis and return a new transform.

        Args:
            m: The shear factor along the x-axis.

        Returns:
            A new transform after shearing.
        """
        return self._compose(
            1,
            m,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
        )

    def shear_y(self, m: float, /) -> Transform2d:
        """Shear along the y-axis and return a new transform.

        Args:
            m: The shear factor along the y-axis.

        Returns:
            A new transform after shearing.
        """
        return self._compose(
            1,
            0,
            0,
            m,
            1,
            0,
            0,
            0,
            1,
        )

    def simple_perspective_x(self, x: float, /) -> Transform2d:
        """Apply perspective along the x-axis with vanishing point at the given x coordinate and return a new transform.

        Args:
            x: The x coordinate of the vanishing point.

        Returns:
            A new transform after applying perspective.
        """
        return self._compose(
            1,
            0,
            0,
            0,
            1,
            0,
            1 / x,
            0,
            1,
        )

    def simple_perspective_y(self, y: float, /) -> Transform2d:
        """Apply perspective along the y-axis with vanishing point at the given y coordinate and return a new transform.

        Args:
            y: The y coordinate of the vanishing point.

        Returns:
            A new transform after applying perspective.
        """
        return self._compose(
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            1 / y,
            1,
        )

    def perspective_x(self, foreground_x: float, vanishing_point: Vec2, /) -> Transform2d:
        """Apply a perspective transformation along the x-axis and return a new transform.

        Args:
            foreground_x: The foreground x-coordinate.
            vanishing_point: The vanishing point vector.

        Returns:
            A new transform after applying perspective.
        """
        return (
            self.simple_perspective_x(vanishing_point.x - foreground_x)
            .shear_y(vanishing_point.y / (vanishing_point.x - foreground_x))
            .translate(Vec2(foreground_x, 0))
        )

    def perspective_y(self, foreground_y: float, vanishing_point: Vec2, /) -> Transform2d:
        """Apply a perspective transformation along the y-axis and return a new transform.

        Args:
            foreground_y: The foreground y-coordinate.
            vanishing_point: The vanishing point vector.

        Returns:
            A new transform after applying perspective.
        """
        return (
            self.simple_perspective_y(vanishing_point.y - foreground_y)
            .shear_x(vanishing_point.x / (vanishing_point.y - foreground_y))
            .translate(Vec2(0, foreground_y))
        )

    def inverse_perspective_x(self, foreground_x: float, vanishing_point: Vec2, /) -> Transform2d:
        """Apply the inverse of a perspective transformation along the x-axis and return a new transform.

        Args:
            foreground_x: The foreground x-coordinate.
            vanishing_point: The vanishing point vector.

        Returns:
            A new transform after applying the inverse perspective.
        """
        return (
            self.translate(Vec2(-foreground_x, 0))
            .shear_y(-vanishing_point.y / (vanishing_point.x - foreground_x))
            .simple_perspective_x(-vanishing_point.x + foreground_x)
        )

    def inverse_perspective_y(self, foreground_y: float, vanishing_point: Vec2, /) -> Transform2d:
        """Apply the inverse of a perspective transformation along the y-axis and return a new transform.

        Args:
            foreground_y: The foreground y-coordinate.
            vanishing_point: The vanishing point vector.

        Returns:
            A new transform after applying the inverse perspective.
        """
        return (
            self.translate(Vec2(0, -foreground_y))
            .shear_x(-vanishing_point.x / (vanishing_point.y - foreground_y))
            .simple_perspective_y(-vanishing_point.y + foreground_y)
        )

    def normalize(self) -> Transform2d:
        """Normalize the transform to have a 1 in the bottom right corner and return a new transform.

        This may fail in some special cases involving perspective transformations where the bottom right corner is 0.

        Returns:
            A new normalized transform.
        """
        assert self.a22 != 0, "Cannot normalize transform with a22 == 0"
        a22 = self.a22 + (self.a22 == 0)
        return Transform2d(
            self.a00 / a22,
            self.a01 / a22,
            self.a02 / a22,
            self.a10 / a22,
            self.a11 / a22,
            self.a12 / a22,
            self.a20 / a22,
            self.a21 / a22,
            1,
        )

    def compose(self, other: Transform2d, /) -> Transform2d:
        """Compose with another transform which is applied after this transform and return a new transform.

        Args:
            other: The other transform to compose with.

        Returns:
            A new transform resulting from the composition.
        """
        return self._compose(
            other.a00,
            other.a01,
            other.a02,
            other.a10,
            other.a11,
            other.a12,
            other.a20,
            other.a21,
            other.a22,
        )

    def compose_before(self, other: Transform2d, /) -> Transform2d:
        """Compose with another transform which is applied before this transform and return a new transform.

        Args:
            other: The other transform to compose with.

        Returns:
            A new transform resulting from the composition.
        """
        return other.compose(self)

    def transform_vec(self, v: Vec2) -> Vec2:
        """Transform a [`Vec2`][sonolus.script.vec.Vec2] and return a new [`Vec2`][sonolus.script.vec.Vec2].

        Args:
            v: The vector to transform.

        Returns:
            A new transformed vector.
        """
        x = self.a00 * v.x + self.a01 * v.y + self.a02
        y = self.a10 * v.x + self.a11 * v.y + self.a12
        w = self.a20 * v.x + self.a21 * v.y + self.a22
        return Vec2(x / w, y / w)

    def transform_quad(self, quad: QuadLike) -> Quad:
        """Transform a [`Quad`][sonolus.script.quad.Quad] and return a new [`Quad`][sonolus.script.quad.Quad].

        Args:
            quad: The quad to transform.

        Returns:
            A new transformed quad.
        """
        return Quad(
            bl=self.transform_vec(quad.bl),
            br=self.transform_vec(quad.br),
            tl=self.transform_vec(quad.tl),
            tr=self.transform_vec(quad.tr),
        )


class InvertibleTransform2d(Record):
    """A transformation matrix for 2D points that can be inverted.

    Usage:
        ```python
        InvertibleTransform2d.new()
        ```
    """

    forward: Transform2d
    inverse: Transform2d

    @classmethod
    def new(cls) -> InvertibleTransform2d:
        """Create a new identity transform.

        Returns:
            A new identity transform.
        """
        return cls(
            forward=Transform2d.new(),
            inverse=Transform2d.new(),
        )

    def translate(self, translation: Vec2, /) -> InvertibleTransform2d:
        """Translate along the x and y axes and return a new transform.

        Args:
            translation: The translation vector.

        Returns:
            A new invertible transform after translation.
        """
        return InvertibleTransform2d(
            forward=self.forward.translate(translation),
            inverse=Transform2d.new().translate(-translation).compose(self.inverse),
        )

    def scale(self, factor: Vec2, /) -> InvertibleTransform2d:
        """Scale about the origin and return a new transform.

        Args:
            factor: The scale factor vector.

        Returns:
            A new invertible transform after scaling.
        """
        return InvertibleTransform2d(
            forward=self.forward.scale(factor),
            inverse=Transform2d.new().scale(Vec2.one() / factor).compose(self.inverse),
        )

    def scale_about(self, factor: Vec2, /, pivot: Vec2) -> InvertibleTransform2d:
        """Scale about the pivot and return a new transform.

        Args:
            factor: The scale factor vector.
            pivot: The pivot point for scaling.

        Returns:
            A new invertible transform after scaling.
        """
        return InvertibleTransform2d(
            forward=self.forward.scale_about(factor, pivot),
            inverse=Transform2d.new().scale_about(Vec2.one() / factor, pivot).compose(self.inverse),
        )

    def rotate(self, angle: float, /) -> InvertibleTransform2d:
        """Rotate about the origin and return a new transform.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.

        Returns:
            A new invertible transform after rotation.
        """
        return InvertibleTransform2d(
            forward=self.forward.rotate(angle),
            inverse=Transform2d.new().rotate(-angle).compose(self.inverse),
        )

    def rotate_about(self, angle: float, /, pivot: Vec2) -> InvertibleTransform2d:
        """Rotate about the pivot and return a new transform.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.
            pivot: The pivot point for rotation.

        Returns:
            A new invertible transform after rotation.
        """
        return InvertibleTransform2d(
            forward=self.forward.rotate_about(angle, pivot),
            inverse=Transform2d.new().rotate_about(-angle, pivot).compose(self.inverse),
        )

    def shear_x(self, m: float, /) -> InvertibleTransform2d:
        """Shear along the x-axis and return a new transform.

        Args:
            m: The shear factor along the x-axis.

        Returns:
            A new invertible transform after shearing.
        """
        return InvertibleTransform2d(
            forward=self.forward.shear_x(m),
            inverse=Transform2d.new().shear_x(-m).compose(self.inverse),
        )

    def shear_y(self, m: float, /) -> InvertibleTransform2d:
        """Shear along the y-axis and return a new transform.

        Args:
            m: The shear factor along the y-axis.

        Returns:
            A new invertible transform after shearing.
        """
        return InvertibleTransform2d(
            forward=self.forward.shear_y(m),
            inverse=Transform2d.new().shear_y(-m).compose(self.inverse),
        )

    def simple_perspective_x(self, x: float, /) -> InvertibleTransform2d:
        """Apply perspective along the x-axis with vanishing point at the given x coordinate and return a new transform.

        Args:
            x: The x coordinate of the vanishing point.

        Returns:
            A new invertible transform after applying perspective.
        """
        return InvertibleTransform2d(
            forward=self.forward.simple_perspective_x(x),
            inverse=Transform2d.new().simple_perspective_x(-x).compose(self.inverse),
        )

    def simple_perspective_y(self, y: float, /) -> InvertibleTransform2d:
        """Apply perspective along the y-axis with vanishing point at the given y coordinate and return a new transform.

        Args:
            y: The y coordinate of the vanishing point.

        Returns:
            A new invertible transform after applying perspective.
        """
        return InvertibleTransform2d(
            forward=self.forward.simple_perspective_y(y),
            inverse=Transform2d.new().simple_perspective_y(-y).compose(self.inverse),
        )

    def perspective_x(self, foreground_x: float, vanishing_point: Vec2, /) -> InvertibleTransform2d:
        """Apply a perspective transformation along the x-axis and return a new transform.

        Args:
            foreground_x: The foreground x-coordinate.
            vanishing_point: The vanishing point vector.

        Returns:
            A new invertible transform after applying perspective.
        """
        return InvertibleTransform2d(
            forward=self.forward.perspective_x(foreground_x, vanishing_point),
            inverse=Transform2d.new().inverse_perspective_x(foreground_x, vanishing_point).compose(self.inverse),
        )

    def perspective_y(self, foreground_y: float, vanishing_point: Vec2, /) -> InvertibleTransform2d:
        """Apply a perspective transformation along the y-axis and return a new transform.

        Args:
            foreground_y: The foreground y-coordinate.
            vanishing_point: The vanishing point vector.

        Returns:
            A new invertible transform after applying perspective.
        """
        return InvertibleTransform2d(
            forward=self.forward.perspective_y(foreground_y, vanishing_point),
            inverse=Transform2d.new().inverse_perspective_y(foreground_y, vanishing_point).compose(self.inverse),
        )

    def normalize(self) -> InvertibleTransform2d:
        """Normalize the transform to have a 1 in the bottom right corner and return a new transform.

        This may fail in some special cases involving perspective transformations where the bottom right corner is 0.

        Returns:
            A new normalized invertible transform.
        """
        return InvertibleTransform2d(
            forward=self.forward.normalize(),
            inverse=self.inverse.normalize(),
        )

    def compose(self, other: InvertibleTransform2d, /) -> InvertibleTransform2d:
        """Compose with another invertible transform which is applied after this transform and return a new transform.

        Args:
            other: The other invertible transform to compose with.

        Returns:
            A new invertible transform resulting from the composition.
        """
        return InvertibleTransform2d(
            forward=self.forward.compose(other.forward),
            inverse=other.inverse.compose(self.inverse),
        )

    def compose_before(self, other: InvertibleTransform2d, /) -> InvertibleTransform2d:
        """Compose with another invertible transform which is applied before this transform and return a new transform.

        Args:
            other: The other invertible transform to compose with.

        Returns:
            A new invertible transform resulting from the composition.
        """
        return other.compose(self)

    def transform_vec(self, v: Vec2) -> Vec2:
        """Transform a [`Vec2`][sonolus.script.vec.Vec2] and return a new [`Vec2`][sonolus.script.vec.Vec2].

        Args:
            v: The vector to transform.

        Returns:
            A new transformed vector.
        """
        return self.forward.transform_vec(v)

    def inverse_transform_vec(self, v: Vec2) -> Vec2:
        """Inverse transform a [`Vec2`][sonolus.script.vec.Vec2] and return a new [`Vec2`][sonolus.script.vec.Vec2].

        Args:
            v: The vector to inverse transform.

        Returns:
            A new inverse transformed vector.
        """
        return self.inverse.transform_vec(v)

    def transform_quad(self, quad: QuadLike) -> Quad:
        """Transform a [`Quad`][sonolus.script.quad.Quad] and return a new [`Quad`][sonolus.script.quad.Quad].

        Args:
            quad: The quad to transform.

        Returns:
            A new transformed quad.
        """
        return self.forward.transform_quad(quad)

    def inverse_transform_quad(self, quad: QuadLike) -> Quad:
        """Inverse transform a [`Quad`][sonolus.script.quad.Quad] and return a new [`Quad`][sonolus.script.quad.Quad].

        Args:
            quad: The quad to inverse transform.

        Returns:
            A new inverse transformed quad.
        """
        return self.inverse.transform_quad(quad)


def perspective_approach(
    distance_ratio: float,
    progress: float,
) -> float:
    """Calculate the perspective correct approach curve given the initial distance, target distance, and progress.

    For typical engines with stage tilt, distance_ratio is the displayed width of a lane at the judge line divided
    by the displayed width of a lane at note spawn. For flat stages, this will be 1.0, and this function would simply
    return progress unchanged.

    Args:
        distance_ratio: The ratio of the distance at note spawn to the distance at the judge line.
        progress: The progress value, where 0 corresponds to note spawn and 1 corresponds to the judge line.

    Returns:
        The perspective-corrected progress value.
    """
    d_0 = distance_ratio
    d_1 = 1.0
    d = max(lerp(d_0, d_1, progress), 1e-6)  # Avoid a zero or negative distance
    return remap(1 / d_0, 1 / d_1, 0, 1, 1 / d)



from pydori.lib.connector import draw_hold_connector, draw_sim_line
from pydori.play.note import Note
from sonolus.script.archetype import EntityRef, PlayArchetype, imported
from sonolus.script.interval import remap
from sonolus.script.runtime import time


class HoldConnector(PlayArchetype):
    """A connector for hold notes."""

    name = "HoldConnector"

    first_ref: EntityRef[Note] = imported()
    second_ref: EntityRef[Note] = imported()

    def should_spawn(self) -> bool:
        return self.first.should_spawn()

    def spawn_order(self) -> float:
        return self.first.spawn_order()

    def update_sequential(self):
        if time() >= self.second.target_time:
            # The target time of the second note has passed, so this connector is no longer needed.
            self.despawn = True
            return
        if self.first.target_time <= time() < self.second.target_time:
            # This connector is the one currently crossing the judgment line,
            # so it has the information to calculate which lane the hold is currently crossing the judgment line at.
            # The hold lane is stored in the note head so the hold manager can use it draw the hold particle and
            # note head at the correct lane.
            self.head.hold_lane = remap(self.first.y, self.second.y, self.first.lane, self.second.lane, 0)

    def update_parallel(self):
        if self.despawn:
            return
        draw_hold_connector(
            self.first.lane,
            self.second.lane,
            self.first.y,
            self.second.y,
        )

    @property
    def first(self):
        return self.first_ref.get()

    @property
    def second(self):
        return self.second_ref.get()

    @property
    def head(self):
        return self.first.head


class SimLine(PlayArchetype):
    """A line connecting two simultaneous notes."""

    name = "SimLine"

    first_ref: EntityRef[Note] = imported()
    second_ref: EntityRef[Note] = imported()

    def should_spawn(self) -> bool:
        return self.first.should_spawn()

    def spawn_order(self) -> float:
        return self.first.spawn_order()

    def update_parallel(self):
        if self.first.is_despawned or self.second.is_despawned:
            self.despawn = True
            return
        draw_sim_line(
            self.first.lane,
            self.second.lane,
            self.first.y,
        )

    @property
    def first(self):
        return self.first_ref.get()

    @property
    def second(self):
        return self.second_ref.get()



from collections.abc import Iterable
from dataclasses import dataclass
from enum import StrEnum
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.ops import Op
from sonolus.script.array_like import ArrayLike, check_positive_index
from sonolus.script.debug import static_error
from sonolus.script.internal.impl import perf_meta_fn
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_function
from sonolus.script.num import Num
from sonolus.script.quad import QuadLike, flatten_quad
from sonolus.script.record import Record
from sonolus.script.vec import Vec2


class Sprite(Record):
    """Skin sprite.

    Usage:
        ```python
        Sprite(id: int)
        ```
    """

    id: int

    @property
    @perf_meta_fn
    def is_available(self) -> bool:
        """Check if the sprite is available."""
        return _has_skin_sprite(self.id)

    @perf_meta_fn
    def draw(self, quad: QuadLike, z: float = 0.0, a: float = 1.0):
        """Draw the sprite.

        Arguments:
            quad: The quad to draw the sprite on.
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw(self.id, *flatten_quad(quad), z, a)

    @perf_meta_fn
    def draw_curved_b(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        """Draw the sprite with a curved bottom with a quadratic Bézier curve.

        Arguments:
            quad: The quad to draw the sprite on.
            cp: The control point of the curve.
            n: The number of segments to approximate the curve (higher is smoother but more expensive).
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw_curved_b(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    @perf_meta_fn
    def draw_curved_t(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        """Draw the sprite with a curved top with a quadratic Bézier curve.

        Arguments:
            quad: The quad to draw the sprite on.
            cp: The control point of the curve.
            n: The number of segments to approximate the curve (higher is smoother but more expensive).
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw_curved_t(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    @perf_meta_fn
    def draw_curved_l(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        """Draw the sprite with a curved left side with a quadratic Bézier curve.

        Arguments:
            quad: The quad to draw the sprite on.
            cp: The control point of the curve.
            n: The number of segments to approximate the curve (higher is smoother but more expensive).
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw_curved_l(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    @perf_meta_fn
    def draw_curved_r(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        """Draw the sprite with a curved right side with a quadratic Bézier curve.

        Arguments:
            quad: The quad to draw the sprite on.
            cp: The control point of the curve.
            n: The number of segments to approximate the curve (higher is smoother but more expensive).
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw_curved_r(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    @perf_meta_fn
    def draw_curved_bt(self, quad: QuadLike, cp1: Vec2, cp2: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        """Draw the sprite with a curved bottom and top with a cubic Bézier curve.

        Arguments:
            quad: The quad to draw the sprite on.
            cp1: The control point of the bottom curve.
            cp2: The control point of the top curve.
            n: The number of segments to approximate the curve (higher is smoother but more expensive).
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw_curved_bt(self.id, *flatten_quad(quad), z, a, n, *cp1.tuple, *cp2.tuple)

    @perf_meta_fn
    def draw_curved_lr(self, quad: QuadLike, cp1: Vec2, cp2: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        """Draw the sprite with a curved left and right side with a cubic Bézier curve.

        Arguments:
            quad: The quad to draw the sprite on.
            cp1: The control point of the left curve.
            cp2: The control point of the right curve.
            n: The number of segments to approximate the curve (higher is smoother but more expensive).
            z: The z-index of the sprite.
            a: The alpha of the sprite.
        """
        _draw_curved_lr(self.id, *flatten_quad(quad), z, a, n, *cp1.tuple, *cp2.tuple)


class SpriteGroup(Record, ArrayLike[Sprite]):
    """A group of sprites.

    Does not support negative indexes.

    Usage:
        ```python
        SpriteGroup(start_id: int, size: int)
        ```
    """

    start_id: int
    size: int

    def __len__(self) -> int:
        return self.size

    def __getitem__(self, index: int) -> Sprite:
        check_positive_index(index, self.size)
        return Sprite(self.start_id + index)

    def get_unchecked(self, index: Num) -> Sprite:
        return Sprite(self.start_id + index)

    def __setitem__(self, index: int, value: Sprite) -> None:
        static_error("SpriteGroup is read-only")


@native_function(Op.HasSkinSprite)
def _has_skin_sprite(sprite_id: int) -> bool:
    raise NotImplementedError


@native_function(Op.Draw)
def _draw(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
) -> None:
    raise NotImplementedError


@native_function(Op.DrawCurvedB)
def _draw_curved_b(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
) -> None:
    raise NotImplementedError


@native_function(Op.DrawCurvedT)
def _draw_curved_t(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
) -> None:
    raise NotImplementedError


@native_function(Op.DrawCurvedL)
def _draw_curved_l(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
) -> None:
    raise NotImplementedError


@native_function(Op.DrawCurvedR)
def _draw_curved_r(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
) -> None:
    raise NotImplementedError


@native_function(Op.DrawCurvedBT)
def _draw_curved_bt(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p1: float,
    q1: float,
    p2: float,
    q2: float,
) -> None:
    raise NotImplementedError


@native_function(Op.DrawCurvedLR)
def _draw_curved_lr(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p1: float,
    q1: float,
    p2: float,
    q2: float,
) -> None:
    raise NotImplementedError


@dataclass
class SkinSprite:
    name: str


@dataclass
class SkinSpriteGroup:
    names: list[str]


def sprite(name: str) -> Any:
    """Define a sprite with the given name."""
    return SkinSprite(name)


def sprite_group(names: Iterable[str]) -> Any:
    """Define a sprite group with the given names."""
    return SkinSpriteGroup(list(names))


type Skin = NewType("Skin", Any)  # type: ignore


class RenderMode(StrEnum):
    """Render mode for sprites."""

    DEFAULT = "default"
    """Use the user's preferred render mode."""

    STANDARD = "standard"
    """Use the standard render mode with bilinear interpolation of textures."""

    LIGHTWEIGHT = "lightweight"
    """Use the lightweight render mode with projective interpolation of textures."""


@dataclass_transform(kw_only_default=True)
def skin[T](cls: type[T]) -> T | Skin:
    """Decorator to define a skin.

    Usage:
        ```python
        @skin
        class Skin:
            render_mode: RenderMode = RenderMode.LIGHTWEIGHT

            note: StandardSprite.NOTE_HEAD_RED
            other: Sprite = sprite("other")
            group_1: SpriteGroup = sprite_group(["one", "two", "three"])
            group_2: SpriteGroup = sprite_group(f"name_{i}" for i in range(10))
        ```
    """
    if len(cls.__bases__) != 1:
        raise ValueError("Skin class must not inherit from any class (except object)")
    instance = cls()
    names = []
    i = 0
    for name, annotation in get_field_specifiers(cls, skip={"render_mode"}).items():
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if len(annotation_values) != 1:
            raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}, too many annotation values")
        sprite_info = annotation_values[0]
        match sprite_info:
            case SkinSprite(name=sprite_name):
                if annotation_type is not Sprite:
                    raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}, expected Sprite")
                names.append(sprite_name)
                setattr(instance, name, Sprite(i))
                i += 1
            case SkinSpriteGroup(names=sprite_names):
                if annotation_type is not SpriteGroup:
                    raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}, expected SpriteGroup")
                start_id = i
                count = len(sprite_names)
                names.extend(sprite_names)
                setattr(instance, name, SpriteGroup(start_id, count))
                i += count
            case _:
                raise TypeError(
                    f"Invalid annotation for skin: {annotation} on field {name}, unknown sprite info, "
                    f"expected a skin() or sprite_group() specifier"
                )
    instance._sprites_ = names
    instance.render_mode = RenderMode(getattr(instance, "render_mode", RenderMode.DEFAULT))
    instance._is_comptime_value_ = True
    return instance


class StandardSprite:
    """Standard skin sprites."""

    NOTE_HEAD_NEUTRAL = Annotated[Sprite, sprite("#NOTE_HEAD_NEUTRAL")]
    NOTE_HEAD_RED = Annotated[Sprite, sprite("#NOTE_HEAD_RED")]
    NOTE_HEAD_GREEN = Annotated[Sprite, sprite("#NOTE_HEAD_GREEN")]
    NOTE_HEAD_BLUE = Annotated[Sprite, sprite("#NOTE_HEAD_BLUE")]
    NOTE_HEAD_YELLOW = Annotated[Sprite, sprite("#NOTE_HEAD_YELLOW")]
    NOTE_HEAD_PURPLE = Annotated[Sprite, sprite("#NOTE_HEAD_PURPLE")]
    NOTE_HEAD_CYAN = Annotated[Sprite, sprite("#NOTE_HEAD_CYAN")]

    NOTE_TICK_NEUTRAL = Annotated[Sprite, sprite("#NOTE_TICK_NEUTRAL")]
    NOTE_TICK_RED = Annotated[Sprite, sprite("#NOTE_TICK_RED")]
    NOTE_TICK_GREEN = Annotated[Sprite, sprite("#NOTE_TICK_GREEN")]
    NOTE_TICK_BLUE = Annotated[Sprite, sprite("#NOTE_TICK_BLUE")]
    NOTE_TICK_YELLOW = Annotated[Sprite, sprite("#NOTE_TICK_YELLOW")]
    NOTE_TICK_PURPLE = Annotated[Sprite, sprite("#NOTE_TICK_PURPLE")]
    NOTE_TICK_CYAN = Annotated[Sprite, sprite("#NOTE_TICK_CYAN")]

    NOTE_TAIL_NEUTRAL = Annotated[Sprite, sprite("#NOTE_TAIL_NEUTRAL")]
    NOTE_TAIL_RED = Annotated[Sprite, sprite("#NOTE_TAIL_RED")]
    NOTE_TAIL_GREEN = Annotated[Sprite, sprite("#NOTE_TAIL_GREEN")]
    NOTE_TAIL_BLUE = Annotated[Sprite, sprite("#NOTE_TAIL_BLUE")]
    NOTE_TAIL_YELLOW = Annotated[Sprite, sprite("#NOTE_TAIL_YELLOW")]
    NOTE_TAIL_PURPLE = Annotated[Sprite, sprite("#NOTE_TAIL_PURPLE")]
    NOTE_TAIL_CYAN = Annotated[Sprite, sprite("#NOTE_TAIL_CYAN")]

    NOTE_CONNECTION_NEUTRAL = Annotated[Sprite, sprite("#NOTE_CONNECTION_NEUTRAL")]
    NOTE_CONNECTION_RED = Annotated[Sprite, sprite("#NOTE_CONNECTION_RED")]
    NOTE_CONNECTION_GREEN = Annotated[Sprite, sprite("#NOTE_CONNECTION_GREEN")]
    NOTE_CONNECTION_BLUE = Annotated[Sprite, sprite("#NOTE_CONNECTION_BLUE")]
    NOTE_CONNECTION_YELLOW = Annotated[Sprite, sprite("#NOTE_CONNECTION_YELLOW")]
    NOTE_CONNECTION_PURPLE = Annotated[Sprite, sprite("#NOTE_CONNECTION_PURPLE")]
    NOTE_CONNECTION_CYAN = Annotated[Sprite, sprite("#NOTE_CONNECTION_CYAN")]

    NOTE_CONNECTION_NEUTRAL_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_NEUTRAL_SEAMLESS")]
    NOTE_CONNECTION_RED_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_RED_SEAMLESS")]
    NOTE_CONNECTION_GREEN_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_GREEN_SEAMLESS")]
    NOTE_CONNECTION_BLUE_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_BLUE_SEAMLESS")]
    NOTE_CONNECTION_YELLOW_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_YELLOW_SEAMLESS")]
    NOTE_CONNECTION_PURPLE_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_PURPLE_SEAMLESS")]
    NOTE_CONNECTION_CYAN_SEAMLESS = Annotated[Sprite, sprite("#NOTE_CONNECTION_CYAN_SEAMLESS")]

    SIMULTANEOUS_CONNECTION_NEUTRAL = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_NEUTRAL")]
    SIMULTANEOUS_CONNECTION_RED = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_RED")]
    SIMULTANEOUS_CONNECTION_GREEN = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_GREEN")]
    SIMULTANEOUS_CONNECTION_BLUE = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_BLUE")]
    SIMULTANEOUS_CONNECTION_YELLOW = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_YELLOW")]
    SIMULTANEOUS_CONNECTION_PURPLE = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_PURPLE")]
    SIMULTANEOUS_CONNECTION_CYAN = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_CYAN")]

    SIMULTANEOUS_CONNECTION_NEUTRAL_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_NEUTRAL_SEAMLESS")]
    SIMULTANEOUS_CONNECTION_RED_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_RED_SEAMLESS")]
    SIMULTANEOUS_CONNECTION_GREEN_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_GREEN_SEAMLESS")]
    SIMULTANEOUS_CONNECTION_BLUE_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_BLUE_SEAMLESS")]
    SIMULTANEOUS_CONNECTION_YELLOW_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_YELLOW_SEAMLESS")]
    SIMULTANEOUS_CONNECTION_PURPLE_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_PURPLE_SEAMLESS")]
    SIMULTANEOUS_CONNECTION_CYAN_SEAMLESS = Annotated[Sprite, sprite("#SIMULTANEOUS_CONNECTION_CYAN_SEAMLESS")]

    DIRECTIONAL_MARKER_NEUTRAL = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_NEUTRAL")]
    DIRECTIONAL_MARKER_RED = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_RED")]
    DIRECTIONAL_MARKER_GREEN = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_GREEN")]
    DIRECTIONAL_MARKER_BLUE = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_BLUE")]
    DIRECTIONAL_MARKER_YELLOW = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_YELLOW")]
    DIRECTIONAL_MARKER_PURPLE = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_PURPLE")]
    DIRECTIONAL_MARKER_CYAN = Annotated[Sprite, sprite("#DIRECTIONAL_MARKER_CYAN")]

    SIMULTANEOUS_MARKER_NEUTRAL = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_NEUTRAL")]
    SIMULTANEOUS_MARKER_RED = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_RED")]
    SIMULTANEOUS_MARKER_GREEN = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_GREEN")]
    SIMULTANEOUS_MARKER_BLUE = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_BLUE")]
    SIMULTANEOUS_MARKER_YELLOW = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_YELLOW")]
    SIMULTANEOUS_MARKER_PURPLE = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_PURPLE")]
    SIMULTANEOUS_MARKER_CYAN = Annotated[Sprite, sprite("#SIMULTANEOUS_MARKER_CYAN")]

    STAGE_MIDDLE = Annotated[Sprite, sprite("#STAGE_MIDDLE")]
    STAGE_LEFT_BORDER = Annotated[Sprite, sprite("#STAGE_LEFT_BORDER")]
    STAGE_RIGHT_BORDER = Annotated[Sprite, sprite("#STAGE_RIGHT_BORDER")]
    STAGE_TOP_BORDER = Annotated[Sprite, sprite("#STAGE_TOP_BORDER")]
    STAGE_BOTTOM_BORDER = Annotated[Sprite, sprite("#STAGE_BOTTOM_BORDER")]

    STAGE_LEFT_BORDER_SEAMLESS = Annotated[Sprite, sprite("#STAGE_LEFT_BORDER_SEAMLESS")]
    STAGE_RIGHT_BORDER_SEAMLESS = Annotated[Sprite, sprite("#STAGE_RIGHT_BORDER_SEAMLESS")]
    STAGE_TOP_BORDER_SEAMLESS = Annotated[Sprite, sprite("#STAGE_TOP_BORDER_SEAMLESS")]
    STAGE_BOTTOM_BORDER_SEAMLESS = Annotated[Sprite, sprite("#STAGE_BOTTOM_BORDER_SEAMLESS")]

    STAGE_TOP_LEFT_CORNER = Annotated[Sprite, sprite("#STAGE_TOP_LEFT_CORNER")]
    STAGE_TOP_RIGHT_CORNER = Annotated[Sprite, sprite("#STAGE_TOP_RIGHT_CORNER")]
    STAGE_BOTTOM_LEFT_CORNER = Annotated[Sprite, sprite("#STAGE_BOTTOM_LEFT_CORNER")]
    STAGE_BOTTOM_RIGHT_CORNER = Annotated[Sprite, sprite("#STAGE_BOTTOM_RIGHT_CORNER")]

    LANE = Annotated[Sprite, sprite("#LANE")]
    LANE_SEAMLESS = Annotated[Sprite, sprite("#LANE_SEAMLESS")]
    LANE_ALTERNATIVE = Annotated[Sprite, sprite("#LANE_ALTERNATIVE")]
    LANE_ALTERNATIVE_SEAMLESS = Annotated[Sprite, sprite("#LANE_ALTERNATIVE_SEAMLESS")]

    JUDGMENT_LINE = Annotated[Sprite, sprite("#JUDGMENT_LINE")]
    NOTE_SLOT = Annotated[Sprite, sprite("#NOTE_SLOT")]
    STAGE_COVER = Annotated[Sprite, sprite("#STAGE_COVER")]

    GRID_NEUTRAL = Annotated[Sprite, sprite("#GRID_NEUTRAL")]
    GRID_RED = Annotated[Sprite, sprite("#GRID_RED")]
    GRID_GREEN = Annotated[Sprite, sprite("#GRID_GREEN")]
    GRID_BLUE = Annotated[Sprite, sprite("#GRID_BLUE")]
    GRID_YELLOW = Annotated[Sprite, sprite("#GRID_YELLOW")]
    GRID_PURPLE = Annotated[Sprite, sprite("#GRID_PURPLE")]
    GRID_CYAN = Annotated[Sprite, sprite("#GRID_CYAN")]


@skin
class EmptySkin:
    pass



from math import isfinite, isinf, isnan

from sonolus.backend.blocks import PlayBlock, PreviewBlock, TutorialBlock, WatchBlock
from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.node import EngineNode, FunctionNode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_reverse_postorder
from sonolus.backend.place import BlockPlace


def cfg_to_engine_node(entry: BasicBlock):
    block_indexes = {block: i for i, block in enumerate(traverse_cfg_reverse_postorder(entry))}
    block_statements = []
    for block in block_indexes:
        statements = []
        statements.extend(ir_to_engine_node(stmt) for stmt in block.statements)
        outgoing = {
            edge.cond: edge.dst for edge in sorted(block.outgoing, key=lambda edge: (edge.cond is None, edge.cond))
        }
        match outgoing:
            case {**other} if not other:
                statements.append(len(block_indexes))
            case {None: target, **other} if not other:
                statements.append(block_indexes[target])
            case {0: f_branch, None: t_branch, **other} if not other:
                statements.append(
                    FunctionNode(
                        func=Op.If,
                        args=(
                            ir_to_engine_node(block.test),
                            block_indexes[t_branch],
                            block_indexes[f_branch],
                        ),
                    )
                )
            case {None: default_branch, **other} if len(other) == 1:
                cond, cond_branch = next(iter(other.items()))
                statements.append(
                    FunctionNode(
                        func=Op.If,
                        args=(
                            ir_to_engine_node(IRPureInstr(Op.Equal, args=[block.test, IRConst(cond)])),
                            block_indexes[cond_branch],
                            block_indexes[default_branch],
                        ),
                    )
                )
            case dict() as targets:
                args = [ir_to_engine_node(block.test)]
                default = len(block_indexes)
                conds = [cond for cond in targets if cond is not None]
                if min(conds) == 0 and max(conds) == len(conds) - 1 and all(int(cond) == cond for cond in conds):
                    args.extend(block_indexes[targets[cond]] for cond in range(len(conds)))
                    if None in targets:
                        default = block_indexes[targets[None]]
                    args.append(default)
                    statements.append(FunctionNode(Op.SwitchIntegerWithDefault, tuple(args)))
                else:
                    for cond, target in targets.items():
                        if cond is None:
                            default = block_indexes[target]
                            continue
                        args.append(cond)
                        args.append(block_indexes[target])
                    args.append(default)
                    statements.append(FunctionNode(Op.SwitchWithDefault, tuple(args)))
        block_statements.append(FunctionNode(Op.Execute, tuple(statements)))
    block_statements.append(0)
    result = FunctionNode(Op.Block, (FunctionNode(Op.JumpLoop, tuple(block_statements)),))
    for block in block_indexes:
        # Clean up without relying on gc
        del block.incoming
        del block.outgoing
        del block.phis
        del block.statements
        del block.test
    return result


def _numeric_to_engine_node(value: float | int) -> EngineNode:
    value = float(value)
    if value.is_integer():
        return int(value)
    elif isfinite(value):
        return value
    elif isinf(value):
        # Read values from ROM
        return FunctionNode(Op.Get, args=(3000, 1 if value > 0 else 2))
    elif isnan(value):
        # Read value from ROM
        return FunctionNode(Op.Get, args=(3000, 0))
    else:
        raise ValueError(f"Invalid constant value: {value}")


def _ir_const_to_engine_node(stmt: IRConst) -> EngineNode:
    return _numeric_to_engine_node(stmt.value)


def _ir_instr_to_engine_node(stmt: IRInstr | IRPureInstr) -> EngineNode:
    return FunctionNode(func=stmt.op, args=tuple(ir_to_engine_node(arg) for arg in stmt.args))


def _ir_get_to_engine_node(stmt: IRGet) -> EngineNode:
    return ir_to_engine_node(stmt.place)


def _block_place_to_engine_node(stmt: BlockPlace) -> EngineNode:
    if stmt.offset == 0:
        index = ir_to_engine_node(stmt.index)
    elif stmt.index == 0:
        index = stmt.offset
    else:
        index = FunctionNode(func=Op.Add, args=(ir_to_engine_node(stmt.index), stmt.offset))
    return FunctionNode(func=Op.Get, args=(ir_to_engine_node(stmt.block), index))


def _ir_set_to_engine_node(stmt: IRSet) -> EngineNode:
    return FunctionNode(func=Op.Set, args=(*ir_to_engine_node(stmt.place).args, ir_to_engine_node(stmt.value)))


_ir_to_engine_node_actions = {
    int: _numeric_to_engine_node,
    PlayBlock: _numeric_to_engine_node,
    WatchBlock: _numeric_to_engine_node,
    PreviewBlock: _numeric_to_engine_node,
    TutorialBlock: _numeric_to_engine_node,
    float: _numeric_to_engine_node,
    IRConst: _ir_const_to_engine_node,
    IRPureInstr: _ir_instr_to_engine_node,
    IRInstr: _ir_instr_to_engine_node,
    IRGet: _ir_get_to_engine_node,
    BlockPlace: _block_place_to_engine_node,
    IRSet: _ir_set_to_engine_node,
}


def ir_to_engine_node(stmt) -> EngineNode:
    return _ir_to_engine_node_actions.get(type(stmt), _block_place_to_engine_node)(stmt)



from pydori.lib.buckets import init_buckets, init_score
from pydori.lib.layout import init_layout
from pydori.lib.stage import (
    draw_stage,
    init_stage_data,
    play_lane_particle,
    schedule_lane_sfx,
)
from pydori.lib.streams import Streams
from pydori.lib.ui import init_ui
from pydori.watch.note import ALL_WATCH_NOTE_TYPES
from sonolus.script.archetype import WatchArchetype, entity_memory
from sonolus.script.runtime import delta_time, is_replay, time
from sonolus.script.timing import time_to_scaled_time


class WatchStage(WatchArchetype):
    """Draws the stage and performs other global game functions."""

    name = "Stage"

    def preprocess(self):
        init_buckets()
        init_score()
        init_ui()
        init_layout()
        init_stage_data()
        for note_type in ALL_WATCH_NOTE_TYPES:
            note_type.global_preprocess()
        self.schedule_effects()

    def spawn_time(self) -> float:
        return -1e8

    def despawn_time(self) -> float:
        return 1e8

    def update_parallel(self):
        draw_stage()

    @staticmethod
    def schedule_effects():
        """Schedule lane sound effects upfront."""
        if not is_replay():
            return
        for effect_time, lanes in Streams.effect_lanes.iter_items_from(-10):
            schedule_lane_sfx(effect_time)
            for lane in lanes:
                WatchScheduledLaneEffect.spawn(time=effect_time, lane=lane)


class WatchScheduledLaneEffect(WatchArchetype):
    """Plays a lane particle effect at a scheduled time."""

    name = "ScheduledLaneEffect"

    time: float = entity_memory()
    lane: float = entity_memory()

    def spawn_time(self) -> float:
        return time_to_scaled_time(self.time)

    def despawn_time(self) -> float:
        return self.spawn_time() + 1

    def update_parallel(self):
        if time() - delta_time() < self.time <= time():
            play_lane_particle(self.lane)



from pydori.level import demo_level
from pydori.lib.options import Options
from pydori.lib.ui import ui_config
from pydori.play.mode import play_mode
from pydori.preview.mode import preview_mode
from pydori.tutorial.mode import tutorial_mode
from pydori.watch.mode import watch_mode
from sonolus.script.engine import Engine, EngineData
from sonolus.script.project import Project

engine = Engine(
    name="pydori",
    title="pydori",
    skin="pixel",
    particle="pixel",
    background="darkblue",
    data=EngineData(
        ui=ui_config,
        options=Options,
        play=play_mode,
        watch=watch_mode,
        preview=preview_mode,
        tutorial=tutorial_mode,
    ),
)

project = Project(
    engine=engine,
    levels=[
        demo_level(),
    ],
)



from typing import Final, Self

from sonolus.script.array_like import ArrayLike, get_positive_index
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.iterator import SonolusIterator
from sonolus.script.maybe import Maybe, Nothing, Some
from sonolus.script.num import Num
from sonolus.script.record import Record


class Range(Record, ArrayLike[int]):
    start: int
    stop: int
    step: int

    def __new__(cls, start: int, stop: int | None = None, step: int = 1):
        if stop is None:
            start, stop = 0, start
        return super().__new__(cls, start, stop, step)

    @classmethod
    @meta_fn
    def frozen(cls, start: int, stop: int | None = None, step: int = 1) -> Self:
        if stop is None:
            start, stop = 0, start
        return super().frozen(start, stop, step)

    def __iter__(self) -> SonolusIterator:
        return RangeIterator(self.start, self.stop, self.step)

    def __contains__(self, item):
        if self.step > 0:
            return self.start <= item < self.stop and (item - self.start) % self.step == 0
        else:
            return self.stop < item <= self.start and (self.start - item) % -self.step == 0

    def __len__(self) -> int:
        if self.step > 0:
            diff = self.stop - self.start
            if diff <= 0:
                return 0
            return (diff + self.step - 1) // self.step
        else:
            diff = self.start - self.stop
            if diff <= 0:
                return 0
            return (diff - self.step - 1) // -self.step

    def __getitem__(self, index: int) -> int:
        return self.start + get_positive_index(index, len(self)) * self.step

    def get_unchecked(self, index: Num) -> int:
        return self.start + index * self.step

    def __setitem__(self, index: int, value: int):
        raise TypeError("Range does not support item assignment")

    @property
    def last(self) -> int:
        return self[len(self) - 1]

    def __eq__(self, other):
        if not isinstance(other, Range):
            return False
        len_self = len(self)
        len_other = len(other)
        if len_self != len_other:
            return False
        if len_self == 0:
            return True
        return self.start == other.start and self.last == other.last

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        raise TypeError("Range is not hashable")


class RangeIterator(Record, SonolusIterator):
    value: int
    stop: Final[int]
    step: Final[int]

    def next(self) -> Maybe[int]:
        has_next = self.value < self.stop if self.step > 0 else self.value > self.stop
        if has_next:
            current = self.value
            self.value += self.step
            return Some(current)
        return Nothing


@meta_fn
def range_or_tuple(start: int, stop: int | None = None, step: int = 1) -> Range | tuple[int, ...]:
    if stop is None:
        start, stop = 0, start
    if not ctx():
        return range(start, stop, step)  # type: ignore
    start = Num._accept_(start)
    stop = Num._accept_(stop) if stop is not None else None
    step = Num._accept_(step)
    if start._is_py_() and stop._is_py_() and step._is_py_():
        start_int = start._as_py_()
        stop_int = stop._as_py_() if stop is not None else None
        if stop_int is None:
            start_int, stop_int = 0, start_int
        step_int = step._as_py_()
        if start_int % 1 != 0 or stop_int % 1 != 0 or step_int % 1 != 0:
            raise TypeError("Range arguments must be integers")
        return validate_value(tuple(range(int(start_int), int(stop_int), int(step_int))))  # type: ignore
    return Range.frozen(start, stop, step)



from sonolus.script.options import options, slider_option, toggle_option
from sonolus.script.text import StandardText


@options
class Options:
    speed: float = slider_option(
        name=StandardText.SPEED,
        standard=True,
        default=1,
        min=0.5,
        max=2,
        step=0.05,
        unit=StandardText.PERCENTAGE_UNIT,
        scope=None,
    )
    note_speed: float = slider_option(
        name=StandardText.NOTE_SPEED,
        default=10,
        min=1,
        max=20,
        step=0.05,
        unit=None,
        scope="pydori",
    )
    note_size: float = slider_option(
        name=StandardText.NOTE_SIZE,
        default=1,
        min=0.1,
        max=2,
        step=0.05,
        unit=StandardText.PERCENTAGE_UNIT,
        scope="pydori",
    )
    lane_width: float = slider_option(
        name=StandardText.LANE_SIZE,
        default=1,
        min=0.1,
        max=1.5,
        step=0.05,
        unit=StandardText.PERCENTAGE_UNIT,
        scope="pydori",
    )
    lane_length: float = slider_option(
        name="Lane Length",
        default=0.8,
        min=0.1,
        max=1,
        step=0.05,
        unit=StandardText.PERCENTAGE_UNIT,
        scope="pydori",
    )

    connector_alpha: float = slider_option(
        name=StandardText.CONNECTOR_ALPHA,
        default=0.8,
        min=0.1,
        max=1,
        step=0.05,
        unit=StandardText.PERCENTAGE_UNIT,
        scope="pydori",
    )

    note_effect_enabled: bool = toggle_option(
        name=StandardText.NOTE_EFFECT,
        default=True,
        scope="pydori",
    )
    lane_effect_enabled: bool = toggle_option(
        name=StandardText.LANE_EFFECT,
        default=True,
        scope="pydori",
    )

    sim_lines_enabled: bool = toggle_option(
        name=StandardText.SIMLINE,
        default=True,
        scope="pydori",
    )
    sim_line_alpha: float = slider_option(
        name=StandardText.SIMLINE_ALPHA,
        default=0.5,
        min=0.1,
        max=1,
        step=0.05,
        unit=StandardText.PERCENTAGE_UNIT,
        scope="pydori",
    )

    sfx_enabled: bool = toggle_option(
        name=StandardText.EFFECT,
        default=True,
        scope="pydori",
    )
    auto_sfx_enabled: bool = toggle_option(
        name=StandardText.EFFECT_AUTO,
        default=False,
        scope="pydori",
    )

    mirror: bool = toggle_option(
        name=StandardText.MIRROR,
        default=False,
        scope="pydori",
    )



from enum import IntEnum

from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function
from sonolus.script.runtime import HorizontalAlign
from sonolus.script.vec import Vec2


class PrintFormat(IntEnum):
    """Print format."""

    NUMBER = 0
    PERCENTAGE = 1
    TIME = 10
    SCORE = 11
    BPM = 20
    TIMESCALE = 21
    BEAT_COUNT = 30
    MEASURE_COUNT = 31
    ENTITY_COUNT = 32


class PrintColor(IntEnum):
    """Print color."""

    THEME = -1
    NEUTRAL = 0
    RED = 1
    GREEN = 2
    BLUE = 3
    YELLOW = 4
    PURPLE = 5
    CYAN = 6


@native_function(Op.Print)
def _print(
    value: int | float,
    format: PrintFormat,  # noqa: A002
    decimal_places: int,
    anchor_x: float,
    anchor_y: float,
    pivot_x: float,
    pivot_y: float,
    width: float,
    height: float,
    rotation: float,
    color: PrintColor,
    alpha: float,
    horizontal_align: HorizontalAlign,
    background: bool,
) -> None:
    raise NotImplementedError


def print_number(
    value: int | float,
    *,
    fmt: PrintFormat,
    decimal_places: int = 0,
    anchor: Vec2,
    pivot: Vec2,
    dimensions: Vec2,
    rotation: float = 0,
    color: PrintColor = PrintColor.THEME,
    alpha: float = 1,
    horizontal_align: HorizontalAlign = HorizontalAlign.LEFT,
    background: bool = False,
):
    """Print a number.

    Only supported in preview mode.

    Args:
        value: The value to print.
        fmt: The print format.
        decimal_places: The number of decimal places.
        anchor: The anchor.
        pivot: The pivot.
        dimensions: The dimensions.
        rotation: The rotation.
        color: The color.
        alpha: The alpha.
        horizontal_align: The horizontal alignment.
        background: Whether to show a background.
    """
    _print(
        value,
        fmt,
        decimal_places,
        anchor.x,
        anchor.y,
        pivot.x,
        pivot.y,
        dimensions.x,
        dimensions.y,
        rotation,
        color,
        alpha,
        horizontal_align,
        background,
    )



import math

from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function


@native_function(Op.Sin, const_eval=True)
def _sin(x: float) -> float:
    return math.sin(x)


@native_function(Op.Cos, const_eval=True)
def _cos(x: float) -> float:
    return math.cos(x)


@native_function(Op.Tan, const_eval=True)
def _tan(x: float) -> float:
    return math.tan(x)


@native_function(Op.Arcsin, const_eval=True)
def _asin(x: float) -> float:
    return math.asin(x)


@native_function(Op.Arccos, const_eval=True)
def _acos(x: float) -> float:
    return math.acos(x)


@native_function(Op.Arctan, const_eval=True)
def _atan(x: float) -> float:
    return math.atan(x)


@native_function(Op.Arctan2, const_eval=True)
def _atan2(y: float, x: float) -> float:
    return math.atan2(y, x)


@native_function(Op.Sinh, const_eval=True)
def _sinh(x: float) -> float:
    return math.sinh(x)


@native_function(Op.Cosh, const_eval=True)
def _cosh(x: float) -> float:
    return math.cosh(x)


@native_function(Op.Tanh, const_eval=True)
def _tanh(x: float) -> float:
    return math.tanh(x)


@native_function(Op.Floor, const_eval=True)
def _floor(x: float) -> float:
    return math.floor(x)


@native_function(Op.Ceil, const_eval=True)
def _ceil(x: float) -> float:
    return math.ceil(x)


@native_function(Op.Trunc, const_eval=True)
def _trunc(x: float) -> float:
    return math.trunc(x)


@native_function(Op.Round, const_eval=True)
def __round(x: float) -> float:
    return round(x)


def _round(x: float, n: int = 0) -> float:
    if n == 0:
        return __round(x)
    return __round(x * 10**n) / 10**n


@native_function(Op.Frac, const_eval=True)
def frac(x: float) -> float:
    return x % 1


@native_function(Op.Log, const_eval=True)
def _ln(x: float) -> float:
    return math.log(x)


def _log(x: float, base: float | None = None) -> float:
    if base is None:
        return _ln(x)
    else:
        return _ln(x) / _ln(base)


def _sqrt(x: float) -> float:
    return x**0.5


@native_function(Op.Degree, const_eval=True)
def _degrees(x: float) -> float:
    """Convert radians to degrees."""
    return math.degrees(x)


@native_function(Op.Radian, const_eval=True)
def _radians(x: float) -> float:
    """Convert degrees to radians."""
    return math.radians(x)


@native_function(Op.Rem, const_eval=True)
def _remainder(x: float, y: float) -> float:
    # This is different from math.remainder in Python's math package, which could be confusing
    return math.copysign(abs(x) % abs(y), x)


MATH_BUILTIN_IMPLS = {
    id(math.sin): _sin,
    id(math.cos): _cos,
    id(math.tan): _tan,
    id(math.asin): _asin,
    id(math.acos): _acos,
    id(math.atan): _atan,
    id(math.atan2): _atan2,
    id(math.sinh): _sinh,
    id(math.cosh): _cosh,
    id(math.tanh): _tanh,
    id(math.floor): _floor,
    id(math.ceil): _ceil,
    id(math.trunc): _trunc,
    id(round): _round,
    id(math.log): _log,
    id(math.sqrt): _sqrt,
}



from math import pi

from pydori.lib.connector import draw_hold_connector
from pydori.lib.layout import lane_to_transformed_vec
from pydori.lib.note import (
    NoteKind,
    draw_note,
    draw_note_head,
    play_note_particle,
    play_note_sfx,
    update_hold_particle,
    update_hold_sfx,
)
from pydori.tutorial.framework import PhaseTime, get_hold_particle, get_hold_sfx, progress_to_y
from pydori.tutorial.instructions import Instructions
from pydori.tutorial.intro import draw_tutorial_intro_note
from pydori.tutorial.painting import (
    paint_hold_flick_motion,
    paint_hold_motion,
    paint_release_motion,
    paint_tap_flick_motion,
    paint_tap_motion,
)
from sonolus.script.array import Array
from sonolus.script.bucket import Judgment
from sonolus.script.interval import remap

ANGLE_RIGHT = 0
ANGLE_UP = pi / 2
ANGLE_LEFT = pi

# Duration of the intro segment where the note intro sprite is shown in the center.
INTRO_DURATION = 1.5

# Duration of the fall phase where the note travels to the judgment line.
FALL_DURATION = 1.5

# Number of times to show the motion during the frozen phase.
FROZEN_REPEATS = 4

# Duration of the tap motion demonstration during each cycle of the frozen phase.
FROZEN_TAP_DURATION = 1

# Duration of the hold motion demonstration during each cycle of the frozen phase.
FROZEN_HOLD_DURATION = 1

# Duration of the flick motion demonstration during each cycle of the frozen phase.
FROZEN_FLICK_DURATION = 0.5

# Duration of the hold release motion demonstration during each cycle of the frozen phase.
FROZEN_RELEASE_DURATION = 1

# Duration after the frozen phase to wait before ending the phase.
END_DURATION = 1.5


def tap_phase(t: PhaseTime):
    intro = t.first(INTRO_DURATION)
    fall = intro.next(FALL_DURATION)
    frozen = fall.next(FROZEN_TAP_DURATION, repeats=FROZEN_REPEATS)
    hit = frozen.end_instant()
    end = frozen.next(END_DURATION)

    if intro:
        draw_tutorial_intro_note(NoteKind.TAP)
    if fall:
        draw_note(
            NoteKind.TAP,
            lane=0,
            y=progress_to_y(fall.progress),
        )
    if frozen:
        draw_note(
            NoteKind.TAP,
            lane=0,
            y=0,
        )
        paint_tap_motion(lane_to_transformed_vec(0), frozen.progress)
        Instructions.tap.show()
    if hit:
        play_note_particle(NoteKind.TAP, 0)
        play_note_sfx(NoteKind.TAP, Judgment.PERFECT)
    if end:
        pass
    return end.is_done


def flick_phase(t: PhaseTime):
    intro = t.first(INTRO_DURATION)
    fall = intro.next(FALL_DURATION)
    frozen = fall.next(FROZEN_TAP_DURATION + FROZEN_FLICK_DURATION, repeats=FROZEN_REPEATS)
    hit = t.instant(frozen.end - FROZEN_FLICK_DURATION)
    end = frozen.next(END_DURATION)

    if intro:
        draw_tutorial_intro_note(NoteKind.FLICK)
    if fall:
        draw_note(
            NoteKind.FLICK,
            lane=0,
            y=progress_to_y(fall.progress),
        )
    if frozen:
        if hit.is_upcoming:
            draw_note(
                NoteKind.FLICK,
                lane=0,
                y=0,
            )
        paint_tap_flick_motion(
            lane_to_transformed_vec(0), ANGLE_UP, frozen.progress, FROZEN_TAP_DURATION, FROZEN_FLICK_DURATION
        )
        Instructions.tap_flick.show()
    if hit:
        play_note_particle(NoteKind.FLICK, 0)
        play_note_sfx(NoteKind.FLICK, Judgment.PERFECT)
    if end:
        pass
    return end.is_done


def directional_flick_phase(t: PhaseTime):
    intro = t.first(INTRO_DURATION)
    fall = intro.next(FALL_DURATION)
    frozen = fall.next(FROZEN_TAP_DURATION + FROZEN_FLICK_DURATION, repeats=FROZEN_REPEATS)
    hit = t.instant(frozen.end - FROZEN_FLICK_DURATION)
    end = frozen.next(END_DURATION)

    if intro:
        draw_tutorial_intro_note(NoteKind.DIRECTIONAL_FLICK, direction=-1, lane=-0.55)
        draw_tutorial_intro_note(NoteKind.DIRECTIONAL_FLICK, direction=1, lane=0.55)
    if fall:
        draw_note(NoteKind.DIRECTIONAL_FLICK, lane=-1, y=progress_to_y(fall.progress), direction=-1)
        draw_note(NoteKind.DIRECTIONAL_FLICK, lane=1, y=progress_to_y(fall.progress), direction=1)
    if frozen:
        if hit.is_upcoming:
            draw_note(NoteKind.DIRECTIONAL_FLICK, lane=-1, y=0, direction=-1)
            draw_note(NoteKind.DIRECTIONAL_FLICK, lane=1, y=0, direction=1)
        paint_tap_flick_motion(
            lane_to_transformed_vec(-1), ANGLE_LEFT, frozen.progress, FROZEN_TAP_DURATION, FROZEN_FLICK_DURATION
        )
        paint_tap_flick_motion(
            lane_to_transformed_vec(1), ANGLE_RIGHT, frozen.progress, FROZEN_TAP_DURATION, FROZEN_FLICK_DURATION
        )
        Instructions.tap_flick.show()
    if hit:
        play_note_particle(NoteKind.DIRECTIONAL_FLICK, lane=-1, direction=-1)
        play_note_particle(NoteKind.DIRECTIONAL_FLICK, lane=1, direction=1)
        play_note_sfx(NoteKind.DIRECTIONAL_FLICK, Judgment.PERFECT)
    if end:
        pass
    return end.is_done


def hold_head_phase(t: PhaseTime):
    intro = t.first(INTRO_DURATION)
    fall = intro.next(FALL_DURATION)
    frozen = fall.next(FROZEN_TAP_DURATION, repeats=FROZEN_REPEATS)
    hit = frozen.end_instant()
    end = frozen.next(END_DURATION)

    if intro:
        draw_tutorial_intro_note(NoteKind.HOLD_HEAD)
    if fall:
        draw_note(
            NoteKind.HOLD_HEAD,
            lane=0,
            y=progress_to_y(fall.progress),
        )
        draw_hold_connector(0, 0, progress_to_y(fall.progress), 99)
    if frozen:
        draw_note(
            NoteKind.HOLD_HEAD,
            lane=0,
            y=0,
        )
        draw_hold_connector(0, 0, 0, 99)
        paint_tap_motion(lane_to_transformed_vec(0), frozen.progress, fade_out=False)
        Instructions.tap_hold.show()
    if hit:
        play_note_particle(NoteKind.HOLD_HEAD, 0)
        play_note_sfx(NoteKind.HOLD_HEAD, Judgment.PERFECT)
    if end:
        particle = get_hold_particle()
        sfx = get_hold_sfx()
        update_hold_particle(particle, 0)
        update_hold_sfx(sfx)
        paint_hold_motion(lane_to_transformed_vec(0))
        draw_hold_connector(0, 0, 0, 99)
        draw_note_head(0)
        Instructions.tap_hold.show()
    return end.is_done


def hold_tick_phase(t: PhaseTime):
    # Lane of each tick.
    tick_lanes = Array(
        0,  # Dummy for the hold head.
        0,
        2,
        0,
        -2,
        0,
        0,  # Dummy for the hold end.
    )

    # The fall progress at which each tick meets the judgment line.
    target_progresses = Array(
        -999,  # Dummy for the hold head.
        0.2,
        0.4,
        0.6,
        0.8,
        1.0,
        999,  # Dummy for the hold end.
    )

    # We do ticks a little differently since we want to show the motion of following ticks.
    # Therefore, the frozen phase comes before the fall phase so we can show the instructions
    # before the ticks fall.
    intro = t.first(INTRO_DURATION)
    frozen = intro.next(FROZEN_HOLD_DURATION, repeats=FROZEN_REPEATS)
    fall = frozen.next(FALL_DURATION)
    end = fall.next(END_DURATION)

    def progress_for_tick(index: int) -> float:
        return remap(target_progresses[index] - 1, target_progresses[index], 0, 1, max(0, fall.progress))

    def tick_y(index: int) -> float:
        return progress_to_y(progress_for_tick(index))

    if intro:
        draw_tutorial_intro_note(NoteKind.HOLD_TICK)

    # Draw the ticks, hold motion, and connectors in all phases after the intro.
    if intro.is_done:
        # Play the particle, draw the head, and paint the hold motion.

        particle = get_hold_particle()
        sfx = get_hold_sfx()

        hold_lane = 0
        for i in range(1, len(tick_lanes)):
            y = tick_y(i)
            prev_y = tick_y(i - 1)
            if y < 0:
                continue
            elif prev_y < 0:
                # Tick i is above the judgment line and tick i-1 is below it,
                # so we can use the two to find the current lane of the hold at the judgment line (y=0)
                hold_lane = remap(prev_y, y, tick_lanes[i - 1], tick_lanes[i], 0)
                break

        update_hold_particle(particle, hold_lane)
        if frozen.is_done:
            # Only play sfx once frozen is done because it's a bit distracting otherwise.
            update_hold_sfx(sfx)
        draw_note_head(hold_lane)
        paint_hold_motion(lane_to_transformed_vec(hold_lane))

        # Draw the connectors.
        for i in range(1, len(tick_lanes)):
            draw_hold_connector(tick_lanes[i - 1], tick_lanes[i], tick_y(i - 1), tick_y(i))

        # Draw ticks and play effects after hit.
        for i, (lane, target_progress) in enumerate(zip(tick_lanes, target_progresses, strict=False)):
            hit_instant = t.instant(fall.start + target_progress * FALL_DURATION)
            if hit_instant.is_upcoming:
                draw_note(
                    NoteKind.HOLD_TICK,
                    lane=lane,
                    y=progress_to_y(progress_for_tick(i)),
                )
            if hit_instant:
                play_note_particle(NoteKind.HOLD_TICK, lane)
                play_note_sfx(NoteKind.HOLD_TICK, Judgment.PERFECT)

    if frozen:
        paint_hold_motion(lane_to_transformed_vec(tick_lanes[0]))
        Instructions.hold_follow.show()
    return end.is_done


def hold_end_phase(t: PhaseTime):
    intro = t.first(INTRO_DURATION)
    fall = intro.next(FALL_DURATION)
    frozen = fall.next(FROZEN_RELEASE_DURATION, repeats=FROZEN_REPEATS)
    hit = frozen.end_instant()
    end = frozen.next(END_DURATION)

    if intro:
        draw_tutorial_intro_note(NoteKind.HOLD_END)
    if fall:
        particle = get_hold_particle()
        sfx = get_hold_sfx()
        draw_note(
            NoteKind.HOLD_END,
            lane=0,
            y=progress_to_y(fall.progress),
        )
        update_hold_particle(particle, 0)
        update_hold_sfx(sfx)
        draw_note_head(0)
        draw_hold_connector(0, 0, 0, progress_to_y(fall.progress))
        paint_hold_motion(lane_to_transformed_vec(0))
    if frozen:
        draw_note(
            NoteKind.HOLD_END,
            lane=0,
            y=0,
        )
        paint_release_motion(lane_to_transformed_vec(0), frozen.progress)
        Instructions.release.show()
    if hit:
        play_note_particle(NoteKind.HOLD_END, 0)
        play_note_sfx(NoteKind.HOLD_END, Judgment.PERFECT)
    return end.is_done


def hold_end_flick_phase(t: PhaseTime):
    intro = t.first(INTRO_DURATION)
    fall = intro.next(FALL_DURATION)
    frozen = fall.next(FROZEN_HOLD_DURATION + FROZEN_FLICK_DURATION, repeats=FROZEN_REPEATS)
    hit = t.instant(frozen.end - FROZEN_FLICK_DURATION)
    end = frozen.next(END_DURATION)

    if intro:
        draw_tutorial_intro_note(NoteKind.FLICK, is_hold_flick_end=True)
    if fall:
        particle = get_hold_particle()
        sfx = get_hold_sfx()
        draw_note(
            NoteKind.FLICK,
            lane=0,
            y=progress_to_y(fall.progress),
        )
        update_hold_particle(particle, 0)
        update_hold_sfx(sfx)
        draw_note_head(0)
        draw_hold_connector(0, 0, 0, progress_to_y(fall.progress))
        paint_hold_motion(lane_to_transformed_vec(0))
    if frozen:
        if hit.is_upcoming:
            draw_note(
                NoteKind.FLICK,
                lane=0,
                y=0,
            )
        paint_hold_flick_motion(
            lane_to_transformed_vec(0), ANGLE_UP, frozen.progress, FROZEN_HOLD_DURATION, FROZEN_FLICK_DURATION
        )
        Instructions.hold_flick.show()
    if hit:
        play_note_particle(NoteKind.FLICK, 0)
        play_note_sfx(NoteKind.FLICK, Judgment.PERFECT)
    return end.is_done


PHASES = (
    tap_phase,
    flick_phase,
    directional_flick_phase,
    hold_head_phase,
    hold_tick_phase,
    hold_end_phase,
    hold_end_flick_phase,
)



from math import pi

from pydori.tutorial.instructions import InstructionIcons
from sonolus.script.interval import interp_clamped, lerp_clamped, remap
from sonolus.script.runtime import runtime_ui
from sonolus.script.vec import Vec2

# Scale factor for flick motion.
FLICK_MOTION_DISTANCE = 0.75


def instruction_scale() -> float:
    return runtime_ui().instruction_config.scale


def instruction_alpha() -> float:
    return runtime_ui().instruction_config.alpha


def _paint_tap(
    pos: Vec2,
    progress: float,
    a: float = 1,
):
    angle = lerp_clamped(pi / 6, pi / 3, progress)
    position = Vec2(0, -1).rotate(pi / 3) * (0.25 * instruction_scale()) + pos
    InstructionIcons.hand.paint(
        position=Vec2(0, 1).rotate(angle) * 0.25 * instruction_scale() + position,
        size=0.25 * instruction_scale(),
        rotation=(180 * angle) / pi,
        z=0,
        a=a * instruction_alpha(),
    )


def paint_tap_motion(pos: Vec2, progress: float, fade_out: bool = True):
    if fade_out:
        a = interp_clamped(
            (0, 0.05, 0.75, 0.95),
            (0, 1, 1, 0),
            progress,
        )
    else:
        a = interp_clamped(
            (0, 0.25),
            (0, 1),
            progress,
        )
    tap_progress = interp_clamped(
        (0.25, 0.75),
        (0, 1),
        progress,
    )
    _paint_tap(pos, tap_progress, a)


def paint_release_motion(pos: Vec2, progress: float):
    a = interp_clamped(
        (0.25, 0.75),
        (1, 0),
        progress,
    )
    tap_progress = interp_clamped(
        (0.25, 0.75),
        (1, 0),  # Reversed for release motion
        progress,
    )
    _paint_tap(
        pos,
        tap_progress,
        a,
    )


def paint_hold_motion(
    pos: Vec2,
    a: float = 1,
):
    _paint_tap(pos, 1, a)


def paint_follow_motion(
    from_pos: Vec2,
    to_pos: Vec2,
    progress: float,
    a: float = 1,
):
    pos = lerp_clamped(from_pos, to_pos, progress)
    paint_hold_motion(pos, a=a)


def paint_flick_motion(
    from_pos: Vec2,
    angle: float,
    progress: float,
):
    to_pos = from_pos + Vec2.unit(angle) * FLICK_MOTION_DISTANCE
    a = interp_clamped(
        (0.25, 0.75),
        (1, 0),
        progress,
    )
    paint_follow_motion(from_pos, to_pos, progress, a=a)


def paint_tap_flick_motion(from_pos: Vec2, angle: float, progress: float, tap_duration: float, flick_duration: float):
    flick_start_progress = tap_duration / (tap_duration + flick_duration)
    if progress < flick_start_progress:
        paint_tap_motion(from_pos, remap(0, flick_start_progress, 0, 1, progress), fade_out=False)
    else:
        paint_flick_motion(from_pos, angle, remap(flick_start_progress, 1, 0, 1, progress))


def paint_hold_flick_motion(
    from_pos: Vec2,
    angle: float,
    progress: float,
    hold_duration: float,
    flick_duration: float,
):
    flick_start_progress = hold_duration / (hold_duration + flick_duration)
    if progress < flick_start_progress:
        paint_hold_motion(from_pos)
    else:
        paint_flick_motion(from_pos, angle, remap(flick_start_progress, 1, 0, 1, progress))



from __future__ import annotations

import textwrap
from collections import deque
from collections.abc import Iterator

from sonolus.backend.ir import IRConst, IRExpr, IRStmt
from sonolus.backend.place import SSAPlace, TempBlock


class FlowEdge:
    src: BasicBlock
    dst: BasicBlock
    cond: float | int | None

    def __init__(self, src: BasicBlock, dst: BasicBlock, cond: float | None = None):
        self.src = src
        self.dst = dst
        self.cond = cond


class BasicBlock:
    phis: dict[SSAPlace | TempBlock, dict[BasicBlock, SSAPlace]]
    statements: list[IRStmt]
    test: IRExpr
    incoming: set[FlowEdge]
    outgoing: set[FlowEdge]

    def __init__(
        self,
        *,
        phi: dict[SSAPlace, dict[BasicBlock, SSAPlace]] | None = None,
        statements: list[IRStmt] | None = None,
        test: IRExpr | None = None,
        incoming: set[FlowEdge] | None = None,
        outgoing: set[FlowEdge] | None = None,
    ):
        self.phis = phi or {}
        self.statements = statements or []
        self.test = test or IRConst(0)
        self.incoming = incoming or set()
        self.outgoing = outgoing or set()

    def connect_to(self, other: BasicBlock, cond: int | float | None = None):
        edge = FlowEdge(self, other, cond)
        self.outgoing.add(edge)
        other.incoming.add(edge)


def traverse_cfg_preorder(block: BasicBlock) -> Iterator[BasicBlock]:
    visited = set()
    queue = deque([block])
    while queue:
        block = queue.popleft()
        if block in visited:
            continue
        visited.add(block)
        yield block
        for edge in sorted(block.outgoing, key=lambda e: (e.cond is None, e.cond)):
            queue.append(edge.dst)


def traverse_cfg_postorder(block: BasicBlock) -> Iterator[BasicBlock]:
    visited = set()

    def dfs(current: BasicBlock):
        if current in visited:
            return
        visited.add(current)
        for edge in sorted(current.outgoing, key=lambda e: (e.cond is None, e.cond)):
            yield from dfs(edge.dst)
        yield current

    yield from dfs(block)


def traverse_cfg_reverse_postorder(block: BasicBlock) -> Iterator[BasicBlock]:
    yield from reversed(list(traverse_cfg_postorder(block)))


def cfg_to_mermaid(entry: BasicBlock):
    def pre(s: str):
        return "\"<pre style='text-align: left;'>" + s.replace("\n", "<br/>") + '</pre>"'

    def fmt(nodes):
        if nodes:
            return "\n".join(str(n) for n in nodes)
        else:
            return "{}"

    block_indexes = {block: i for i, block in enumerate(traverse_cfg_reverse_postorder(entry))}

    lines = ["Entry([Entry]) --> 0"]
    for block, index in block_indexes.items():
        lines.append(
            f"{index}[{
                pre(
                    fmt(
                        [
                            f'#{index}',
                            *(
                                f'{dst} := phi({
                                    ", ".join(
                                        f"{block_indexes.get(src_block, '<dead>')}: {src_place}"
                                        for src_block, src_place in sorted(
                                            phis.items(), key=lambda x: block_indexes.get(x[0])
                                        )
                                    )
                                })'
                                for dst, phis in block.phis.items()
                            ),
                            *block.statements,
                        ]
                    )
                )
            }]"
        )

        outgoing = {edge.cond: edge.dst for edge in block.outgoing}
        match outgoing:
            case {**other} if not other:
                lines.append(f"{index} --> Exit")
            case {None: target, **other} if not other:
                lines.append(f"{index} --> {block_indexes[target]}")
            case {0: f_branch, None: t_branch, **other} if not other:
                lines.append(f"{index}_{{{{{pre(fmt([block.test]))}}}}}")
                lines.append(f"{index} --> {index}_")
                lines.append(f"{index}_ --> |true| {block_indexes[t_branch]}")
                lines.append(f"{index}_ --> |false| {block_indexes[f_branch]}")
            case dict() as tgt:
                lines.append(f"{index}_{{{{{pre(fmt([block.test]))}}}}}")
                lines.append(f"{index} --> {index}_")
                for cond, target in sorted(tgt.items(), key=lambda x: (x[0] is None, x[0])):
                    lines.append(
                        f"{index}_ --> |{pre(fmt([cond if cond is not None else 'default']))}| {block_indexes[target]}"
                    )
    lines.append("Exit([Exit])")

    body = textwrap.indent("\n".join(lines), "    ")
    return f"graph\n{body}"


def hash_cfg(entry: BasicBlock) -> int:
    # Note: not stable between Python runs due to hash randomization
    block_indexes = {block: i for i, block in enumerate(traverse_cfg_preorder(entry))}

    h = 0
    for block, index in block_indexes.items():
        outgoing = [(edge.cond, block_indexes.get(edge.dst, -1)) for edge in block.outgoing]
        h = hash(
            (h, index, tuple(block.statements), block.test, tuple(sorted(outgoing, key=lambda x: (x[0] is None, x[0]))))
        )

    return h


def cfg_to_text(entry: BasicBlock) -> str:
    def indent(iterable, prefix="  "):
        for line in iterable:
            yield f"{prefix}{line}"

    block_indexes = {block: i for i, block in enumerate(traverse_cfg_reverse_postorder(entry))}

    def format_phis(phis):
        for dst, phi_srcs in phis.items():
            srcs = ", ".join(
                f"{block_indexes.get(src_block, '<dead>')}: {src_place}"
                for src_block, src_place in sorted(phi_srcs.items(), key=lambda x: block_indexes.get(x[0]))
            )
            yield f"{dst} := phi({srcs})\n"

    def format_statements(statements):
        for stmt in statements:
            yield f"{stmt}\n"

    def format_outgoing(outgoing_edges, test, indexes):
        outgoing = {edge.cond: edge.dst for edge in outgoing_edges}
        match outgoing:
            case {**other} if not other:
                yield "goto exit\n"
            case {None: target, **other} if not other:
                yield f"goto {indexes[target]}\n"
            case {0: f_branch, None: t_branch, **other} if not other:
                yield f"goto {indexes[t_branch]} if {test} else {indexes[f_branch]}\n"
            case dict() as tgt:
                yield f"goto when {test}\n"
                yield from indent(
                    f"{('default' if cond is None else str(cond))} -> {indexes[target]}\n"
                    for cond, target in sorted(tgt.items(), key=lambda x: (x[0] is None, x[0]))
                )

    def format_blocks():
        for block, index in block_indexes.items():
            yield f"{index}:\n"
            yield from indent(format_phis(block.phis))
            yield from indent(format_statements(block.statements))
            yield from indent(format_outgoing(block.outgoing, block.test, block_indexes))

    return "".join(format_blocks())



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from tests.script.conftest import run_and_validate


def test_dict_get_existing():
    def fn():
        d = {"a": 1, "b": 2}
        return Array(d["a"], d["b"])

    assert run_and_validate(fn) == Array(1, 2)


def test_dict_in_operator():
    def fn():
        d = {"a": 1, "b": 2}
        return Array(
            "a" in d,  # True
            "b" in d,  # True
            "c" in d,  # False
            "d" in d,  # False
            "a" not in d,  # False
            "b" not in d,  # False
            "c" not in d,  # True
            "d" not in d,  # True
        )

    assert run_and_validate(fn) == Array(True, True, False, False, False, False, True, True)


@given(
    v1=st.integers(min_value=-10, max_value=10),
    v2=st.integers(min_value=-10, max_value=10),
)
def test_dict_union(v1, v2):
    def fn():
        d1 = {"a": v1}
        d2 = {"b": v2}
        d3 = d1 | d2
        return Array(d3["a"], d3["b"])

    assert run_and_validate(fn) == Array(v1, v2)



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.num import Num
from sonolus.script.values import alloc, copy, sizeof, swap, zeros
from sonolus.script.vec import Vec2
from tests.script.conftest import run_and_validate

floats = st.floats(min_value=-1000, max_value=1000, allow_nan=False, allow_infinity=False)


def test_alloc_num():
    def fn():
        # The returned value is actually pretty useless since Num is immutable
        alloc(Num)
        return 1

    assert run_and_validate(fn) == 1


def test_alloc_vec2():
    def fn():
        v = alloc(Vec2)
        v @= Vec2(1, 2)
        return v

    assert run_and_validate(fn) == Vec2(1, 2)


def test_alloc_array():
    def fn():
        arr = alloc(Array[Num, 3])
        arr[0] = 1
        arr[1] = 2
        arr[2] = 3
        return arr

    result = run_and_validate(fn)
    assert result[0] == 1
    assert result[1] == 2
    assert result[2] == 3


def test_zeros_num():
    def fn():
        return zeros(Num)

    assert run_and_validate(fn) == 0


def test_zeros_vec2():
    def fn():
        return zeros(Vec2)

    assert run_and_validate(fn) == Vec2(0, 0)


def test_zeros_array():
    def fn():
        return zeros(Array[Num, 3])

    result = run_and_validate(fn)
    assert result[0] == 0
    assert result[1] == 0
    assert result[2] == 0


@given(value=floats)
def test_copy_num(value):
    def fn():
        original = value
        copied = copy(original)
        return Array(original, copied)

    result = run_and_validate(fn)
    assert result[0] == value
    assert result[1] == value


def test_copy_vec2():
    def fn():
        original = Vec2(3, 4)
        copied = copy(original)
        return Array(original.x, original.y, copied.x, copied.y)

    result = run_and_validate(fn)
    assert result[0] == 3
    assert result[1] == 4
    assert result[2] == 3
    assert result[3] == 4


def test_copy_array():
    def fn():
        original = Array(1, 2, 3)
        copied = copy(original)
        original[0] = 99
        return Array(original[0], copied[0])

    result = run_and_validate(fn)
    assert result[0] == 99
    assert result[1] == 1


def test_copy_independence():
    def fn():
        original = Vec2(1, 2)
        copied = copy(original)
        copied @= Vec2(99, 2)
        return Array(original.x, copied.x)

    result = run_and_validate(fn)
    assert result[0] == 1
    assert result[1] == 99


def test_swap_vec2():
    def fn():
        v1 = Vec2(1, 2)
        v2 = Vec2(3, 4)
        swap(v1, v2)
        return Array(v1.x, v1.y, v2.x, v2.y)

    result = run_and_validate(fn)
    assert result[0] == 3
    assert result[1] == 4
    assert result[2] == 1
    assert result[3] == 2


def test_swap_array():
    def fn():
        arr1 = Array(1, 2, 3)
        arr2 = Array(4, 5, 6)
        swap(arr1, arr2)
        return Array(arr1[0], arr1[1], arr1[2], arr2[0], arr2[1], arr2[2])

    result = run_and_validate(fn)
    assert result[0] == 4
    assert result[1] == 5
    assert result[2] == 6
    assert result[3] == 1
    assert result[4] == 2
    assert result[5] == 3


def test_sizeof_num():
    def fn():
        return sizeof(Num)

    result = run_and_validate(fn)
    assert result == 1


def test_sizeof_vec2():
    def fn():
        return sizeof(Vec2)

    result = run_and_validate(fn)
    assert result == 2


def test_sizeof_array():
    def fn():
        return sizeof(Array[Num, 5])

    result = run_and_validate(fn)
    assert result == 5
